import glob
import os
from copy import deepcopy

import numpy as np
import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import check_if_filepath_list_is_empty
from simba.utils.errors import (AnnotationFileNotFoundError,
                                ColumnNotFoundError,
                                ThirdPartyAnnotationEventCountError,
                                ThirdPartyAnnotationOverlapError)
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import get_fn_ext, read_df, write_df
from simba.utils.warnings import (
    ThirdPartyAnnotationsClfMissingWarning,
    ThirdPartyAnnotationsOutsidePoseEstimationDataWarning)

TIME_FIELD = "Time_Relative_hmsf"
VIDEO_NAME_FIELD = "Observation"
BEHAVIOR_FIELD = "Behavior"
EVENT_TYPE_FIELD = "Event_Type"
POINT_EVENT = "Point"
START = "State start"
STOP = "State stop"
EXPECTED_FIELDS = [TIME_FIELD, VIDEO_NAME_FIELD, BEHAVIOR_FIELD, EVENT_TYPE_FIELD]


class NoldusObserverImporter(ConfigReader):
    """
    Append Noldus Observer human annotations onto featurized pose-estimation data.
    Results are saved within the project_folder/csv/targets_inserted directory of
    the SimBA project (as parquets' or CSVs).

    :param str config_path: path to SimBA project config file in Configparser format
    :param str data_dir: path to folder holding Observer data files is XLSX or XLS format

    .. note::
       `Third-party import GitHub tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/third_party_annot.md>`__.
       `Expected input example 1 <https://github.com/sgoldenlab/simba/blob/master/misc/Observer_example_1.xlsx>`__.
       `Expected input example 2 <https://github.com/sgoldenlab/simba/blob/master/misc/Observer_example_2.xlsx>`__.

    Examples
    -----
    >>> _ = NoldusObserverImporter(config_path='MyConfigPath', data_dir='MyNoldusObserverDataDir').run()
    """

    def __init__(self, config_path: str, data_dir: str):
        super().__init__(config_path=config_path)
        self.observer_files_found = glob.glob(data_dir + "/*.xlsx") + glob.glob(
            data_dir + "/*.xls"
        )
        self.observer_files_found = [
            x for x in self.observer_files_found if "~$" not in x
        ]
        check_if_filepath_list_is_empty(
            filepaths=self.observer_files_found,
            error_msg=f"SIMBA ERROR: The {data_dir} directory contains ZERO xlsx/xls files",
        )
        check_if_filepath_list_is_empty(
            filepaths=self.feature_file_paths,
            error_msg=f"SIMBA ERROR: The {self.features_dir} directory contains ZERO files",
        )
        self.__read_data()

    def __check_column_names(self, df: pd.DataFrame, file_path: str):
        remain = list(set(EXPECTED_FIELDS) - set(list(df.columns)))
        if remain:
            raise ColumnNotFoundError(file_name=file_path, column_name=remain[0])

    def check_timestamps(self, timestamps=list):
        corrected_ts = []
        for timestamp in timestamps:
            h, m, s = timestamp.split(":", 3)
            missing_fractions = 9 - len(s)
            if missing_fractions == 0:
                corrected_ts.append(timestamp)
            else:
                corrected_ts.append(f'{h}:{m}:{s}.{"0" * missing_fractions}')
        return corrected_ts

    def __read_data(self):
        print(
            f"Reading Noldus Observer annotation files ({str(len(self.observer_files_found))} files)..."
        )
        self.annotation = {}
        for file_path in self.observer_files_found:
            try:
                df = pd.read_excel(
                    file_path, sheet_name=None, usecols=EXPECTED_FIELDS
                ).popitem(last=False)[1]
            except KeyError:
                raise ColumnNotFoundError(
                    file_name=file_path, column_name=", ".join(EXPECTED_FIELDS)
                )
            for video_name in df[VIDEO_NAME_FIELD].unique():
                video_df = df[df[VIDEO_NAME_FIELD] == video_name].reset_index(drop=True)
                self.__check_column_names(df=video_df, file_path=file_path)
                video_df = video_df[video_df[EVENT_TYPE_FIELD] != POINT_EVENT]
                video_df[TIME_FIELD] = self.check_timestamps(
                    timestamps=list(video_df[TIME_FIELD].astype(str))
                )
                video_df[TIME_FIELD] = pd.to_timedelta(video_df[TIME_FIELD])
                video_df[EVENT_TYPE_FIELD] = video_df[EVENT_TYPE_FIELD].replace(
                    {START: "START", STOP: "STOP"}
                )
                _, _, fps = self.read_video_info(video_name=video_name)
                video_df["FRAME"] = video_df[TIME_FIELD].dt.total_seconds() * fps
                video_df["FRAME"] = video_df["FRAME"].apply(np.floor)
                video_df = video_df.drop([TIME_FIELD, VIDEO_NAME_FIELD], axis=1)
                if video_name in list(self.annotation.keys()):
                    self.annotation[video_name] = pd.concat(
                        [self.annotation[video_name], video_df], axis=0
                    ).reset_index(drop=True)
                else:
                    self.annotation[video_name] = video_df
        for k, v in self.annotation.items():
            self.annotation[k] = v.sort_values(by="FRAME").reset_index(drop=True)
        print(
            f"Annotations for {str(len(list(self.annotation.keys())))} video names found in Ethovision files..."
        )

    def run(self):
        for file_path in self.feature_file_paths:
            video_timer = SimbaTimer()
            video_timer.start_timer()
            _, file_name, _ = get_fn_ext(filepath=file_path)
            if file_name not in list(self.annotation.keys()):
                raise AnnotationFileNotFoundError(video_name=file_name)
            data_df = read_df(file_path=file_path, file_type=self.file_type)
            output_df = deepcopy(data_df)
            for clf_name in self.clf_names:
                clf_df = self.annotation[file_name][[EVENT_TYPE_FIELD, "FRAME"]][
                    self.annotation[file_name][BEHAVIOR_FIELD] == clf_name
                ].reset_index(drop=True)
                start_events, stop_events = len(
                    clf_df[clf_df[EVENT_TYPE_FIELD] == "START"]
                ), len(clf_df[clf_df[EVENT_TYPE_FIELD] == "STOP"])
                if start_events != stop_events:
                    raise ThirdPartyAnnotationEventCountError(
                        video_name=file_name,
                        clf_name=clf_name,
                        start_event_cnt=start_events,
                        stop_event_cnt=stop_events,
                    )
                start_df, stop_df = clf_df[
                    clf_df[EVENT_TYPE_FIELD] == "START"
                ].reset_index(drop=True), clf_df[
                    clf_df[EVENT_TYPE_FIELD] == "STOP"
                ].reset_index(
                    drop=True
                )
                start_df, stop_df = start_df["FRAME"].rename(
                    columns={"FRAME": "START"}
                ), stop_df["FRAME"].rename(columns={"FRAME": "STOP"})
                clf_df = pd.concat([start_df, stop_df], axis=1).reset_index(drop=True)
                clf_df.columns = ["START", "STOP"]
                if len(clf_df.query("START > STOP")) > 0:
                    raise ThirdPartyAnnotationOverlapError(
                        video_name=file_name, clf_name=clf_name
                    )
                if len(clf_df) == 0:
                    ThirdPartyAnnotationsClfMissingWarning(
                        video_name=file_name, clf_name=clf_name
                    )
                    output_df[clf_name] = 0
                    continue
                annot_idx = list(
                    clf_df.apply(
                        lambda x: list(range(int(x["START"]), int(x["STOP"]) + 1)), 1
                    )
                )
                annot_idx = [x for xs in annot_idx for x in xs]
                idx_diff = list(set(annot_idx) - set(data_df.index))
                if len(idx_diff) > 0:
                    ThirdPartyAnnotationsOutsidePoseEstimationDataWarning(
                        video_name=file_name,
                        clf_name=clf_name,
                        frm_cnt=data_df.index[-1],
                        first_error_frm=idx_diff[0],
                        ambiguous_cnt=len(idx_diff),
                    )
                    annot_idx = [x for x in annot_idx if x not in idx_diff]
                output_df[clf_name] = 0
                output_df.loc[annot_idx, clf_name] = 1
            self.__save(
                df=output_df,
                path=os.path.join(
                    self.targets_folder, file_name + "." + self.file_type
                ),
            )
            video_timer.stop_timer()
            print(
                f"Imported Noldus Observer annotations for video {file_name} (elapsed time {video_timer.elapsed_time_str}s)..."
            )
        self.timer.stop_timer()
        stdout_success(
            msg=f"Imported annotations saved in project/folder/csv/targets_inserted directory",
            elapsed_time=self.timer.elapsed_time_str,
        )

    def __save(self, df: pd.DataFrame, path: str):
        write_df(df=df, file_type=self.file_type, save_path=path)


# test = NoldusObserverImporter(config_path='/Users/simon/Desktop/envs/troubleshooting/Gosia/project_folder/project_config.ini',
#                               data_dir='/Users/simon/Desktop/envs/troubleshooting/Gosia/source/behaviours/Exp_38')
# test.run()


# for k, v in test.annotation.items():
#     print(v[BEHAVIOR_FIELD].unique())
#
#
#
# # test.run()



from tkinter import *

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.roi_tools.ROI_analyzer import ROIAnalyzer
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, DropDownMenu,
                                        Entry_Box)
from simba.utils.checks import check_float
from simba.utils.enums import ConfigKey, Formats, Keys, Links


class ROIAnalysisPopUp(ConfigReader, PopUpMixin):
    def __init__(self, config_path: str):
        ConfigReader.__init__(self, config_path=config_path)
        PopUpMixin.__init__(self, title="ROI ANALYSIS", size=(400, 400))
        self.animal_cnt_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="SELECT NUMBER OF ANIMALS",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.ROI_DATA_ANALYSIS.value,
        )
        self.animal_cnt_dropdown = DropDownMenu(
            self.animal_cnt_frm,
            "# of animals",
            list(range(1, self.animal_cnt + 1)),
            labelwidth=20,
        )
        self.animal_cnt_dropdown.setChoices(1)
        self.animal_cnt_confirm_btn = Button(
            self.animal_cnt_frm,
            text="Confirm",
            command=lambda: self.create_settings_frm(),
        )
        self.animal_cnt_frm.grid(row=0, column=0, sticky=NW)
        self.animal_cnt_dropdown.grid(row=0, column=0, sticky=NW)
        self.animal_cnt_confirm_btn.grid(row=0, column=1, sticky=NW)
        self.main_frm.mainloop()

    def create_settings_frm(self):
        if hasattr(self, "setting_frm"):
            self.setting_frm.destroy()
            self.body_part_frm.destroy()

        self.setting_frm = LabelFrame(
            self.main_frm, text="SETTINGS", font=Formats.LABELFRAME_HEADER_FORMAT.value
        )
        self.choose_bp_frm(parent=self.setting_frm, bp_options=self.body_parts_lst)
        self.choose_bp_threshold_frm(parent=self.setting_frm)
        self.calculate_distances_frm = LabelFrame(
            self.setting_frm,
            text="CALCULATE DISTANCES",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
        )
        self.calculate_distance_moved_var = BooleanVar(value=False)
        self.calculate_distance_moved_cb = Checkbutton(
            self.calculate_distances_frm,
            text="Compute distances moved within ROIs",
            variable=self.calculate_distance_moved_var,
        )
        self.calculate_distance_moved_cb.grid(row=0, column=0, sticky=NW)
        self.detailed_roi_frm = LabelFrame(
            self.setting_frm,
            text="DETAILED ROI BOUT DATA",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
        )
        self.detailed_roi_var = BooleanVar(value=False)
        self.detailed_roi_cb = Checkbutton(
            self.detailed_roi_frm,
            text="Save detailed ROI bout data",
            variable=self.detailed_roi_var,
        )
        self.detailed_roi_cb.grid(row=1, column=0, sticky=NW)
        self.calculate_distances_frm.grid(
            row=self.frame_children(frame=self.setting_frm), column=0, sticky=NW
        )
        self.detailed_roi_frm.grid(
            row=self.frame_children(frame=self.setting_frm) + 1, column=0, sticky=NW
        )
        self.setting_frm.grid(row=1, column=0, sticky=NW)
        self.create_run_frm(run_function=self.run)

        self.main_frm.mainloop()

    def run(self):
        settings = {}
        check_float(
            name="Probability threshold",
            value=self.probability_entry.entry_get,
            min_value=0.00,
            max_value=1.00,
        )
        settings["threshold"] = float(self.probability_entry.entry_get)
        settings["body_parts"] = {}
        self.config.set(
            ConfigKey.ROI_SETTINGS.value,
            ConfigKey.ROI_ANIMAL_CNT.value,
            str(self.animal_cnt_dropdown.getChoices()),
        )
        self.config.set(
            ConfigKey.ROI_SETTINGS.value,
            ConfigKey.PROBABILITY_THRESHOLD.value,
            str(self.probability_entry.entry_get),
        )
        for cnt, dropdown in self.body_parts_dropdowns.items():
            settings["body_parts"][f"animal_{cnt + 1}_bp"] = dropdown.getChoices()
            self.config.set(
                "ROI settings", f"animal_{cnt + 1}_bp", str(dropdown.getChoices())
            )
        self.update_config()
        roi_analyzer = ROIAnalyzer(
            ini_path=self.config_path,
            data_path="outlier_corrected_movement_location",
            calculate_distances=self.calculate_distance_moved_var.get(),
            detailed_bout_data=self.detailed_roi_var.get(),
            settings=settings,
        )
        roi_analyzer.run()
        roi_analyzer.save()
        self.root.destroy()


# ROIAnalysisPopUp(config_path='/Users/simon/Desktop/envs/simba_dev/tests/data/test_projects/mouse_open_field/project_folder/project_config.ini')

# ROIAnalysisPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/Termites_5/project_folder/project_config.ini')



__author__ = "Simon Nilsson"

from tkinter import *

from simba.data_processors.severity_bout_based_calculator import \
    SeverityBoutCalculator
from simba.data_processors.severity_frame_based_calculator import \
    SeverityFrameCalculator
from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import CreateLabelFrameWithIcon, DropDownMenu
from simba.utils.enums import Formats, Keys, Links
from simba.utils.errors import NoSpecifiedOutputError


class AnalyzeSeverityPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str):
        PopUpMixin.__init__(self, title="SIMBA SEVERITY ANALYSIS")
        ConfigReader.__init__(self, config_path=config_path)

        if len(self.multi_animal_id_list) > 1:
            self.multi_animal_id_list.insert(0, "ALL ANIMALS")
        self.clip_cnt_options = list(range(1, 20))
        self.clip_cnt_options.insert(0, "ALL CLIPS")
        self.frame_cnt_var = BooleanVar(value=False)
        self.seconds_cnt_var = BooleanVar(value=False)
        self.settings_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="SETTINGS",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.ANALYZE_ML_RESULTS.value,
        )
        self.clf_dropdown = DropDownMenu(
            self.settings_frm, "CLASSIFIER:", self.clf_names, "25"
        )
        self.clf_dropdown.setChoices(self.clf_names[0])
        self.brackets_dropdown = DropDownMenu(
            self.settings_frm, "BRACKETS:", list(range(1, 21)), "25"
        )
        self.bracket_type_dropdown = DropDownMenu(
            self.settings_frm, "BRACKET TYPE: ", ["QUANTIZE", "QUANTILE"], "25"
        )
        self.brackets_dropdown.setChoices(10)
        self.bracket_type_dropdown.setChoices("QUANTIZE")
        self.animal_dropdown = DropDownMenu(
            self.settings_frm, "ANIMALS", self.multi_animal_id_list, "25"
        )
        self.animal_dropdown.setChoices(self.multi_animal_id_list[0])
        self.bouts_vs_frames_dropdown = DropDownMenu(
            self.settings_frm,
            "DATA TYPE: ",
            ["BOUTS", "FRAMES"],
            "25",
            com=self.show_settings,
        )
        self.bouts_vs_frames_dropdown.setChoices("FRAMES")
        self.normalization_type_dropdown = DropDownMenu(
            self.settings_frm,
            "MOVEMENT NORMALIZATION TYPE: ",
            ["ALL VIDEOS", "SINGLE VIDEOS"],
            "25",
        )
        self.normalization_type_dropdown.setChoices("ALL VIDEOS")
        self.save_bracket_cut_off_points_var = BooleanVar(value=True)
        self.savet_brackets_info_cb = Checkbutton(
            self.settings_frm,
            text="Save bracket definitions",
            variable=self.save_bracket_cut_off_points_var,
        )
        self.further_settings_frm = LabelFrame(
            self.main_frm,
            text="Further settings",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            pady=5,
            padx=5,
            fg="black",
        )
        run_frm = LabelFrame(
            self.main_frm,
            text="RUN",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            pady=5,
            padx=5,
            fg="black",
        )
        run_btn = Button(
            run_frm, text="RUN SEVERITY ANALYSIS", command=lambda: self.run()
        )
        self.settings_frm.grid(row=0, column=0, sticky=NW)
        self.clf_dropdown.grid(row=0, column=0, sticky=NW)
        self.brackets_dropdown.grid(row=1, column=0, sticky=NW)
        self.bracket_type_dropdown.grid(row=2, column=0, sticky=NW)
        self.animal_dropdown.grid(row=3, column=0, sticky=NW)
        self.bouts_vs_frames_dropdown.grid(row=4, column=0, sticky=NW)
        self.normalization_type_dropdown.grid(row=5, column=0, sticky=NW)
        self.savet_brackets_info_cb.grid(row=6, column=0, sticky=NW)
        self.further_settings_frm.grid(row=2, column=0, sticky=NW)
        run_frm.grid(row=3, column=0, sticky=NW)
        run_btn.grid(row=0, column=0, sticky=NW)
        self.show_settings(self.bouts_vs_frames_dropdown.getChoices())
        self.main_frm.mainloop()

    def show_settings(self, setting):
        for w in self.further_settings_frm.winfo_children():
            w.destroy()

        self.further_settings_frm.grid()
        self.visualize_var = BooleanVar(value=False)
        self.show_pose_var = BooleanVar(value=False)
        self.visualize_cb = Checkbutton(
            self.further_settings_frm,
            text="Visualize",
            variable=self.visualize_var,
            command=self.enable_visualization_options,
        )
        self.show_pose_cb = Checkbutton(
            self.further_settings_frm,
            text="Show pose-estimated locations",
            variable=self.show_pose_var,
        )
        self.video_speed_dropdown = DropDownMenu(
            self.further_settings_frm,
            "Video speed:",
            [float(i) / 10 for i in range(1, 21)],
            "25",
        )
        self.video_speed_dropdown.setChoices(1.0)
        self.number_of_clips_dropdown = DropDownMenu(
            self.further_settings_frm, "Clip count:", self.clip_cnt_options, "25"
        )
        self.show_pose_cb.config(state=DISABLED)
        self.video_speed_dropdown.disable()
        self.number_of_clips_dropdown.disable()
        self.visualize_cb.grid(row=0, column=0, sticky=NW)
        self.show_pose_cb.grid(row=1, column=0, sticky=NW)
        self.video_speed_dropdown.grid(row=2, column=0, sticky=NW)
        self.number_of_clips_dropdown.grid(row=3, column=0, sticky=NW)
        self.number_of_clips_dropdown.setChoices("ALL CLIPS")

        if setting == "FRAMES":
            self.frame_cnt_cb = Checkbutton(
                self.further_settings_frm,
                text="FRAME COUNT",
                variable=self.frame_cnt_var,
            )
            self.seconds_cnt_cb = Checkbutton(
                self.further_settings_frm, text="SECONDS", variable=self.seconds_cnt_var
            )
            self.frame_cnt_cb.grid(row=4, column=0, sticky=NW)
            self.seconds_cnt_cb.grid(row=5, column=0, sticky=NW)

    def enable_visualization_options(self):
        if self.visualize_var.get():
            self.show_pose_cb.config(state=NORMAL)
            self.video_speed_dropdown.enable()
            self.number_of_clips_dropdown.enable()
        else:
            self.show_pose_cb.config(state=DISABLED)
            self.video_speed_dropdown.disable()
            self.number_of_clips_dropdown.disable()

    def run(self):
        if self.animal_dropdown.getChoices() == "ALL ANIMALS":
            animals = self.multi_animal_id_list[1:]
        else:
            animals = [self.animal_dropdown.getChoices()]
        settings = {
            "brackets": int(self.brackets_dropdown.getChoices()),
            "bracket_type": self.bracket_type_dropdown.getChoices(),
            "clf": self.clf_dropdown.getChoices(),
            "animals": animals,
            "save_bin_definitions": self.save_bracket_cut_off_points_var.get(),
            "normalization": self.normalization_type_dropdown.getChoices(),
        }
        settings["visualize"] = self.visualize_var.get()
        settings["video_speed"] = float(self.video_speed_dropdown.getChoices())
        settings["visualize_event_cnt"] = self.number_of_clips_dropdown.getChoices()
        settings["show_pose"] = self.show_pose_var.get()
        if self.bouts_vs_frames_dropdown.getChoices() == "FRAMES":
            settings["frames"] = self.frame_cnt_var.get()
            settings["time"] = self.seconds_cnt_var.get()
            if (not self.seconds_cnt_var.get()) and (not self.frame_cnt_var.get()):
                raise NoSpecifiedOutputError(
                    msg="SIMBA ERROR: Please select frames and/or time output metrics",
                    source=self.__class__.__name__,
                )

        if self.bouts_vs_frames_dropdown.getChoices() == "FRAMES":
            severity_calculator = SeverityFrameCalculator(
                config_path=self.config_path, settings=settings
            )
        else:
            severity_calculator = SeverityBoutCalculator(
                config_path=self.config_path, settings=settings
            )

        severity_calculator.run()


# _ = AnalyzeSeverityPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')



import pytest
import os

from simba.outlier_tools.outlier_corrector_location import OutlierCorrecterLocation
from simba.outlier_tools.outlier_corrector_movement import OutlierCorrecterMovement
from simba.outlier_tools.skip_outlier_correction import OutlierCorrectionSkipper

@pytest.mark.parametrize("config_path", ['tests/data/test_projects/two_c57/project_folder/project_config.ini'])
def test_outlier_corrector_movement_use_case(config_path):
    movement_corrector = OutlierCorrecterMovement(config_path=config_path)
    movement_corrector.run()

@pytest.mark.parametrize("config_path", ['tests/data/test_projects/two_c57/project_folder/project_config.ini'])
def test_outlier_corrector_location_use_case(config_path):
    location_corrector = OutlierCorrecterLocation(config_path=config_path)
    location_corrector.run()

@pytest.mark.parametrize("config_path", ['tests/data/test_projects/two_c57/project_folder/project_config.ini'])
def test_outlier_skipper_use_case(config_path):
    skipper = OutlierCorrectionSkipper(config_path=config_path)
    skipper.run()




import glob
import os
from typing import Tuple, Union

import pandas as pd

try:
    from typing import Literal
except:
    from typing_extensions import Literal

from simba.utils.checks import check_float, check_str
from simba.utils.enums import ConfigKey, Dtypes, Paths
from simba.utils.lookups import get_bp_config_code_class_pairs
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (get_fn_ext, read_config_entry,
                                    read_config_file,
                                    read_project_path_and_file_type,
                                    read_video_info_csv)


def feature_extraction_runner(config_path: Union[str, os.PathLike]) -> None:
    """
    Helper to run feature extraction from CLI.

    :param config_path: Path to SimBA project config file in ini format.
    """

    config = read_config_file(config_path=config_path)
    pose_setting = read_config_entry(
        config=config,
        section=ConfigKey.CREATE_ENSEMBLE_SETTINGS.value,
        option=ConfigKey.POSE_SETTING.value,
        data_type=Dtypes.STR.value,
    )
    animal_cnt = read_config_entry(
        config=config,
        section=ConfigKey.GENERAL_SETTINGS.value,
        option=ConfigKey.ANIMAL_CNT.value,
        data_type=Dtypes.INT.value,
    )
    print(
        f"Pose-estimation body part setting for feature extraction: {animal_cnt} animals {animal_cnt} body-parts"
    )
    feature_extractor_classes = get_bp_config_code_class_pairs()
    if pose_setting == "8":
        feature_extractor = feature_extractor_classes[pose_setting][animal_cnt](
            config_path=config_path
        )
    else:
        feature_extractor = feature_extractor_classes[pose_setting](
            config_path=config_path
        )
    feature_extractor.run()


def set_video_parameters(
    config_path: Union[str, os.PathLike],
    px_per_mm: float,
    fps: float,
    resolution: Tuple[int, int],
) -> None:
    """
    Helper to batch set the video_info.csv from CLI. Requires all videos to have the same pixels per millimeter,
    fps, and resolution.

    :param str config_path: Path to SimBA project config file in ini format.
    :param float px_per_mm: Pixels per millimeter in all the video files.
    :param int fps: FPS of the video files.
    :param tuple resolution: Resolution of the video files.
    """

    timer = SimbaTimer(start=True)
    config = read_config_file(config_path=config_path)
    project_path, file_type = read_project_path_and_file_type(config=config)
    video_info_path = os.path.join(project_path, Paths.VIDEO_INFO.value)
    if not os.path.isfile(video_info_path):
        df = pd.DataFrame(
            columns=[
                "Video",
                "fps",
                "Resolution_width",
                "Resolution_height",
                "Distance_in_mm",
                "pixels/mm",
            ]
        ).set_index("Video")
        df.to_csv(video_info_path)
    video_info = read_video_info_csv(os.path.join(project_path, Paths.VIDEO_INFO.value))
    data_paths = glob.glob(
        os.path.join(project_path, Paths.OUTLIER_CORRECTED.value) + "/*." + file_type
    )
    for file_path in data_paths:
        _, video_name, _ = get_fn_ext(file_path)
        if video_name not in list(video_info["Video"]):
            video_info.loc[len(video_info)] = [
                video_name,
                fps,
                resolution[0],
                resolution[1],
                0,
                px_per_mm,
            ]
    video_info.reset_index(drop=True).set_index("Video").to_csv(
        os.path.join(project_path, Paths.VIDEO_INFO.value)
    )
    timer.stop_timer()
    stdout_success(
        msg="Video parameters set",
        elapsed_time=timer.elapsed_time_str,
        source=set_video_parameters.__name__,
    )


def set_outlier_correction_criteria_cli(
    config_path: Union[str, os.PathLike],
    movement_criterion: float,
    location_criterion: float,
    aggregation: Literal["mean", "median"],
    body_parts: dict,
):
    """Helper to set outlier settings in a SimBA project_config.ini from command line"""
    timer = SimbaTimer(start=True)
    config = read_config_file(config_path=config_path)
    check_float(name="MOVEMENT CRITERION", value=movement_criterion, min_value=0.1)
    check_float(name="LOCATION CRITERION", value=location_criterion, min_value=0.1)
    check_str(name="AGGREGATION METHOD", value=aggregation, options=("mean", "median"))
    config.set(
        section="Outlier settings",
        option="movement_criterion",
        value=str(movement_criterion),
    )
    config.set(
        section="Outlier settings",
        option="location_criterion",
        value=str(location_criterion),
    )
    config.set(section="Outlier settings", option="mean_or_median", value=aggregation)
    for cnt, (k, v) in enumerate(body_parts.items()):
        config.set(
            section="Outlier settings",
            option=f"movement_bodypart1_mouse{cnt+1}",
            value=v["Movement"][0],
        )
        config.set(
            section="Outlier settings",
            option=f"movement_bodypart2_mouse{cnt+1}",
            value=v["Movement"][1],
        )
        config.set(
            section="Outlier settings",
            option=f"location_bodypart1_mouse{cnt+1}",
            value=v["Location"][0],
        )
        config.set(
            section="Outlier settings",
            option=f"location_bodypart2_mouse{cnt+1}",
            value=v["Location"][1],
        )
    with open(config_path, "w") as file:
        config.write(file)

    timer.stop_timer()
    stdout_success(
        "Outlier parameters set",
        elapsed_time=timer.elapsed_time_str,
        source=set_outlier_correction_criteria_cli.__name__,
    )


# DEFINITIONS
# from simba.outlier_tools.outlier_corrector_movement import OutlierCorrecterMovement
# from simba.outlier_tools.outlier_corrector_location import OutlierCorrecterLocation
#
# CONFIG_PATH = '/Users/simon/Desktop/envs/troubleshooting/DLC_2_Black_animals/project_folder/project_config.ini'
# AGGREGATION_METHOD = 'mean'
# BODY_PARTS = {'Animal_1': {'Movement': ['Nose_1', 'Tail_base_1'],
#                            'Location': ['Nose_1', 'Tail_base_1']},
#               'Animal_2': {'Movement': ['Nose_2', 'Tail_base_2'],
#                            'Location': ['Nose_2', 'Tail_base_2']}}
# MOVEMENT_CRITERION = 0.7
# LOCATION_CRITERION = 2.0
#
# set_outlier_correction_criteria_cli(config_path=CONFIG_PATH,
#                                     aggregation=AGGREGATION_METHOD,
#                                     body_parts=BODY_PARTS,
#                                     movement_criterion=MOVEMENT_CRITERION,
#                                     location_criterion=LOCATION_CRITERION)
#
#
# _ = OutlierCorrecterMovement(config_path=CONFIG_PATH).run()
# _ = OutlierCorrecterLocation(config_path=CONFIG_PATH).run()


# set_video_parameters(config_path='/Users/simon/Desktop/envs/troubleshooting/notebook_example/project_folder/project_config.ini', px_per_mm=5.6, fps=25, resolution=(400, 400))



import numpy as np

def sliding_circular_mean(data: np.ndarray, time_windows: np.ndarray, fps: int) -> np.ndarray:
    """
    Compute the circular mean in degrees within sliding temporal windows.

    :parameter np.ndarray data: 1d array with feature values in degrees.
    :parameter np.ndarray time_windows: Rolling time-windows as floats in seconds. E.g., [0.2, 0.4, 0.6]
    :parameter int fps: fps of the recorded video
    :returns np.ndarray: Size data.shape[0] x time_windows.shape[0] array

    .. image:: _static/img/mean_rolling_timeseries_angle.png
       :width: 600
       :align: center

    .. attention::
       The returned values represents the angular mean dispersion in the time-window ``[current_frame-time_window->current_frame]``.
       `-1` is returned when ``current_frame-time_window`` is less than 0.

    :example:
    >>> data = np.random.normal(loc=45, scale=1, size=20).astype(np.float32)
    >>> CircularStatisticsMixin().sliding_circular_mean(data=data,time_windows=np.array([0.5, 1.0]), fps=10)
    """

    data = np.deg2rad(data)
    results = np.full((data.shape[0], time_windows.shape[0]), -1.0)
    for time_window in range(time_windows.shape[0]):
        window_size = int(time_windows[time_window] * fps)
        for current_frm in range(window_size-1, results.shape[0]):
            data_window = data[(current_frm - window_size)+1: current_frm+1]
            m = np.rad2deg(np.arctan2(np.mean(np.sin(data_window)), np.mean(np.cos(data_window))))
            results[current_frm, time_window] = np.abs(np.round(m, 4))

    return results

data = np.array([0, 10, 20, 30, 40, 50, 60, 70, 80]).astype(np.float32)


x = sliding_circular_mean(data=data, time_windows=np.array([1.0]), fps=3)




__author__ = "Simon Nilsson"

import time
from typing import List, Tuple

import numpy as np
from numba import jit, njit, prange, typed

from simba.utils.data import fast_mean_rank


class CircularStatisticsMixin(object):
    """
    Mixin for circular statistics. Unlike linear data, circular data wrap around in a circular or periodic
    manner such as two measurements of e.g., 360 vs. 1 are more similar than two measurements of 1 vs. 3. Thus, the
    minimum and maximum values are connected, forming a closed loop, and we therefore need specialized
    statistical methods.

    These methods have support for multiple animals and base radial directions derived from two or three body-parts.

    Methods are adopted from the referenced packages below which are **far** more reliable. However,
    runtime on standard hardware (multicore CPU) is prioritized and typically orders of magnitude faster than referenced libraries.

    See image below for example of expected run-times for a small set of method examples included in this class.

    .. note::
        Many method has numba typed `signatures <https://numba.pydata.org/numba-doc/latest/reference/types.html>`_ to decrease
        compilation time. Make sure to pass the correct dtypes as indicated by signature decorators.

    .. important::
        See references below for mature packages computing more extensive circular measurements.

    .. image:: _static/img/circular_statistics.png
       :width: 800
       :align: center

    .. image:: _static/img/circular_stats_runtimes.png
       :width: 1200
       :align: center

    References
    ----------
    .. [1] `pycircstat <https://github.com/circstat/pycircstat>`_.
    .. [2] `circstat <https://www.mathworks.com/matlabcentral/fileexchange/10676-circular-statistics-toolbox-directional-statistics>`_.
    .. [3] `pingouin.circular <https://github.com/raphaelvallat/pingouin/blob/master/pingouin/circular.py>`_.
    .. [4] `pycircular <https://github.com/albahnsen/pycircular>`_.
    .. [5] `scipy.stats.directional_stats <https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.directional_stats.html>`_.
    .. [6] `astropy.stats.circstats <https://docs.astropy.org/en/stable/_modules/astropy/stats/circstats.html>`_.
    """

    def __init__(self):
        pass

    @staticmethod
    @njit("(float32[:],)")
    def mean_resultant_vector_length(data: np.ndarray) -> float:
        """
        Jitted compute of the mean resultant vector length of a single sample. Captures the overall "pull" or "tendency" of the
        data points towards a central direction on the circle with a range between 0 and 1.

        .. image:: _static/img/mean_resultant_vector.png
           :width: 600
           :align: center

        .. math::

            R = \\frac{{\\sqrt{{\\sum_{{i=1}}^N \\cos(\\theta_i - \\bar{\theta})^2 + \\sum_{{i=1}}^N \\sin(\theta_i - \\bar{\\theta})^2}}}}{{N}}

        where \(N\) is the number of data points, \(\theta_i\) is the angle of the ith data point, and \(\bar{\theta}\) is the mean angle.

        :parameter np.ndarray data: 1D array of size len(frames) representing angles in degrees.
        :returns float: The mean resultant vector of the angles. 1 represents tendency towards a single point. 0 represents no central point.

        :example:
        >>> data = np.array([50, 90, 70, 60, 20, 90]).astype(np.float32)
        >>> CircularStatisticsMixin().mean_resultant_vector_length(data=data)
        >>> 0.9132277170817057
        """

        data = np.deg2rad(data)
        mean_angles = np.arctan2(np.mean(np.sin(data)), np.mean(np.cos(data)))
        return np.sqrt(
            np.sum(np.cos(data - mean_angles)) ** 2
            + np.sum(np.sin(data - mean_angles)) ** 2
        ) / len(data)

    @staticmethod
    @njit("(float32[:], float64, float64[:])")
    def sliding_mean_resultant_vector_length(
        data: np.ndarray, fps: int, time_windows: np.ndarray
    ) -> np.ndarray:
        """
        Jitted compute of the mean resultant vector within sliding time window. Captures the overall "pull" or "tendency" of the
        data points towards a central direction on the circle with a range between 0 and 1.

        .. attention::
           The returned values represents resultant vector length in the time-window ``[(current_frame-time_window)->current_frame]``.
           `-1` is returned where ``current_frame-time_window`` is less than 0.

        .. image:: _static/img/sliding_mean_resultant_length.png
           :width: 600
           :align: center

        :parameter np.ndarray data: 1D array of size len(frames) representing degrees.
        :parameter np.ndarray time_window: Rolling time-window as float in seconds.
        :parameter int fps: fps of the recorded video
        :returns np.ndarray: Size len(data) x len(time_windows) representing resultant vector length in the prior ``time_window``.

        :example:
        >>> data_1, data_2 = np.random.normal(loc=45, scale=1, size=100), np.random.normal(loc=90, scale=45, size=100)
        >>> data = np.hstack([data_1, data_2])
        >>> CircularStatisticsMixin().sliding_mean_resultant_vector_length(data=data.astype(np.float32),time_windows=np.array([1.0]), fps=10)

        """
        data = np.deg2rad(data)
        results = np.full((data.shape[0], time_windows.shape[0]), -1.0)
        for time_window_cnt in prange(time_windows.shape[0]):
            window_size = int(time_windows[time_window_cnt] * fps)
            for window_end in prange(window_size, data.shape[0] + 1, 1):
                window_data = data[window_end - window_size : window_end]
                mean_angles = np.arctan2(
                    np.mean(np.sin(window_data)), np.mean(np.cos(window_data))
                )
                r = np.sqrt(
                    np.sum(np.cos(window_data - mean_angles)) ** 2
                    + np.sum(np.sin(window_data - mean_angles)) ** 2
                ) / len(window_data)
                results[window_end - 1] = r
        return results

    @staticmethod
    @njit("(float32[:],)")
    def circular_mean(data: np.ndarray) -> float:
        """
        Jitted compute of the circular mean of single sample.

        :parameter np.ndarray data: 1D array of size len(frames) representing angles in degrees.
        :returns float: The circular mean of the angles in degrees.

        .. image:: _static/img/mean_angle.png
           :width: 400
           :align: center

        :example:
        >>> data = np.array([50, 90, 70, 60, 20, 90]).astype(np.float32)
        >>> CircularStatisticsMixin().circular_mean(data=data)
        >>> 63.737892150878906
        """

        m = np.rad2deg(np.arctan2(np.mean(np.sin(np.deg2rad(data))), np.mean(np.cos(np.deg2rad(data)))))
        return np.abs(np.round(m, 4))

    @staticmethod
    @njit("(float32[:], float64[:], float64)")
    def sliding_circular_mean(data: np.ndarray, time_windows: np.ndarray, fps: int) -> np.ndarray:
        """
        Compute the circular mean in degrees within sliding temporal windows.

        :parameter np.ndarray data: 1d array with feature values in degrees.
        :parameter np.ndarray time_windows: Rolling time-windows as floats in seconds. E.g., [0.2, 0.4, 0.6]
        :parameter int fps: fps of the recorded video
        :returns np.ndarray: Size data.shape[0] x time_windows.shape[0] array

        .. image:: _static/img/mean_rolling_timeseries_angle.png
           :width: 600
           :align: center

        .. attention::
           The returned values represents the angular mean dispersion in the time-window ``[current_frame-time_window->current_frame]``.
           `-1` is returned when ``current_frame-time_window`` is less than 0.

        :example:
        >>> data = np.random.normal(loc=45, scale=1, size=20).astype(np.float32)
        >>> CircularStatisticsMixin().sliding_circular_mean(data=data,time_windows=np.array([0.5, 1.0]), fps=10)
        """

        data = np.deg2rad(data)
        results = np.full((data.shape[0], time_windows.shape[0]), -1.0)
        for time_window in range(time_windows.shape[0]):
            window_size = int(time_windows[time_window] * fps)
            for current_frm in range(window_size - 1, results.shape[0]):
                data_window = data[(current_frm - window_size) + 1: current_frm + 1]
                m = np.rad2deg(np.arctan2(np.mean(np.sin(data_window)), np.mean(np.cos(data_window))))
                results[current_frm, time_window] = np.abs(np.round(m, 4))
        return results

    @staticmethod
    @njit("(float32[:],)")
    def circular_std(data: np.ndarray) -> float:
        """
        Jitted compute of the circular standard deviation from a single distribution of angles in degrees.

        :parameter ndarray data: 1D array of size len(frames) with angles in degrees
        :returns float: The standard deviation of the data sample in degrees

        .. image:: _static/img/circular_std.png
           :width: 600
           :align: center

        .. math::

           \\sigma_{\\text{circular}} = \\text{rad2deg}\\left(\\sqrt{-2 \\cdot \\log\\left(|\text{mean}(\\exp(j \\cdot \\theta))|\\right)}\\right)

        where \\(\\theta\\) represents the angles in radians

        :example:
        >>> data = np.array([180, 221, 32, 42, 212, 101, 139, 41, 69, 171, 149, 200]).astype(np.float32)
        >>> CircularStatisticsMixin().circular_std(data=data)
        >>> 75.03725024504664
        """

        data = np.deg2rad(data)
        return np.abs(np.rad2deg(np.sqrt(-2 * np.log(np.abs(np.mean(np.exp(1j * data)))))))

    @staticmethod
    @njit("(float32[:], int64, float64[:])")
    def sliding_circular_std(data: np.ndarray, fps: int, time_windows: np.ndarray) -> np.ndarray:
        """
        Compute standard deviation of angular data in sliding time windows.

        .. image:: _static/img/angle_stdev.png
           :width: 600
           :align: center

        :parameter ndarray data: 1D array of size len(frames) representing degrees.
        :parameter np.ndarray time_window: Sliding time-window as float in seconds.
        :parameter int fps: fps of the recorded video
        :returns np.ndarray: Size data.shape[0] x time_windows.shape[0] with angular standard deviations in rolling time windows in degrees.

        :example:
        >>> data = np.array([180, 221, 32, 42, 212, 101, 139, 41, 69, 171, 149, 200]).astype(np.float32)
        >>> CircularStatisticsMixin().sliding_circular_std(data=data.astype(np.float32), time_windows=np.array([0.5]), fps=10)
        """

        data = np.deg2rad(data)
        results = np.full((data.shape[0], time_windows.shape[0]), 0.0)
        for time_window_cnt in prange(time_windows.shape[0]):
            window_size = int(time_windows[time_window_cnt] * fps)
            for window_end in prange(window_size, data.shape[0] + 1, 1):
                window_data = data[window_end - window_size : window_end]
                m = np.rad2deg(np.sqrt(-2 * np.log(np.abs(np.mean(np.exp(1j * window_data))))))
                results[window_end - 1][time_window_cnt] = np.abs(np.round(m, 4))

        return results.astype(np.float32)

    @staticmethod
    @njit("(float32[:], int64)")
    def instantaneous_angular_velocity(data: np.ndarray, bin_size: int) -> np.ndarray:
        """
        Jitted compute of absolute angular change in the smallest possible time bin.

        .. note::
            If the smallest possible frame-to-frame time-bin in Video 1 is 33ms (recorded at 30fps), and the
            smallest possible frame-to-frame time-bin in Video 2 is 66ms (recorded at 15fps), we correct for
            this across recordings using the ``bin_size`` argument. E.g., when passing angular data from Video 1
            we pass bin_size as ``2``, and when passing angular data for Video 2 we set bin_size to ``1`` to
            allow comparisons of instantaneous angular velocity between Video 1 and Video 2.

            When current frame minus bin_size results in a negative index, -1 is returned.

        .. image:: _static/img/instantaneous_angular_velocity.png
           :width: 600
           :align: center

        :parameter ndarray data: 1D array of size len(frames) representing degrees.
        :parameter int bin_size: The number of frames prior to compare the current angular velocity against.

        :example:
        >>> data = np.array([350, 360, 365, 360]).astype(np.float32)
        >>> CircularStatisticsMixin().instantaneous_angular_velocity(data=data, bin_size=1.0)
        >>> [-1., 10.00002532, 4.999999, 4.999999]
        >>> CircularStatisticsMixin().instantaneous_angular_velocity(data=data, bin_size=2)
        >>> [-1., -1., 15.00002432, 0.]
        """
        data = np.deg2rad(data)
        results = np.full((data.shape[0]), -1)
        left_idx, right_idx = 0, bin_size
        for end_idx in prange(right_idx, data.shape[0] + 1, 1):
            v = np.rad2deg(np.pi - np.abs(np.pi - np.abs(data[left_idx] - data[end_idx])))
            results[end_idx] = int(np.round(v, 4))
            left_idx += 1
        return results

    @staticmethod
    @njit("(float32[:],)")
    def degrees_to_cardinal(data: np.ndarray) -> List[str]:
        """
        Convert degree angles to cardinal direction bucket e.g., 0 -> "N", 180 -> "S"

        .. note::
           To convert cardinal literals to integers, map using ``simba.utils.enums.lookups.cardinality_to_integer_lookup``.
           To convert integers to cardinal literals, map using ``simba.utils.enums.lookups.integer_to_cardinality_lookup``.

        .. image:: _static/img/degrees_to_cardinal.png
           :width: 600
           :align: center

        :parameter np.ndarray degree_angles: 1d array of degrees. Note: return by ``self.head_direction``.
        :return List[str]: List of strings representing frame-wise cardinality.

        :example:
        >>> data = np.array(list(range(0, 405, 45))).astype(np.float32)
        >>> CircularStatisticsMixin().degrees_to_cardinal(degree_angles=data)
        >>> ['N', 'NE', 'E', 'SE', 'S', 'SW', 'W', 'NW', 'N']
        """

        DIRECTIONS = ["N", "NE", "E", "SE", "S", "SW", "W", "NW"]
        results = typed.List(["str"])
        for i in prange(data.shape[0]):
            ix = round(data[i] / (360.0 / len(DIRECTIONS)))
            direction = DIRECTIONS[ix % len(DIRECTIONS)]
            results.append(direction)
        return results[1:]

    @staticmethod
    @njit("(float32[:,:], float32[:, :], float32[:, :])")
    def direction_three_bps(
        nose_loc: np.ndarray, left_ear_loc: np.ndarray, right_ear_loc: np.ndarray
    ) -> np.ndarray:
        """
        Jitted helper to compute the degree angle from three body-parts. Computes the angle in degrees left_ear <-> nose
        and right_ear_nose and returns the midpoint.

        .. image:: _static/img/angle_from_3_bps.png
          :width: 600
          :align: center

        :parameter ndarray nose_loc: 2D array of size len(frames)x2 representing nose coordinates
        :parameter ndarray left_ear_loc: 2D array of size len(frames)x2 representing left ear coordinates
        :parameter ndarray right_ear_loc: 2D array of size len(frames)x2 representing right ear coordinates
        :return np.ndarray: Array of size nose_loc.shape[0] with direction in degrees.

        :example:
        >>> nose_loc = np.random.randint(low=0, high=500, size=(50, 2)).astype(np.float32)
        >>> left_ear_loc = np.random.randint(low=0, high=500, size=(50, 2)).astype(np.float32)
        >>> right_ear_loc = np.random.randint(low=0, high=500, size=(50, 2)).astype(np.float32)
        >>> results = CircularStatisticsMixin().direction_three_bps(nose_loc=nose_loc, left_ear_loc=left_ear_loc, right_ear_loc=right_ear_loc)
        """

        results = np.full((nose_loc.shape[0]), np.nan)
        for i in prange(nose_loc.shape[0]):
            left_ear_to_nose = np.arctan2(
                nose_loc[i][0] - left_ear_loc[i][0], left_ear_loc[i][1] - nose_loc[i][1]
            )
            right_ear_nose = np.arctan2(
                nose_loc[i][0] - right_ear_loc[i][0],
                right_ear_loc[i][1] - nose_loc[i][1],
            )
            mean_angle_rad = np.arctan2(
                np.sin(left_ear_to_nose) + np.sin(right_ear_nose),
                np.cos(left_ear_to_nose) + np.cos(right_ear_nose),
            )
            results[i] = (np.degrees(mean_angle_rad) + 360) % 360
        return results

    @staticmethod
    @njit("(float32[:, :], float32[:, :])")
    def direction_two_bps(
        anterior_loc: np.ndarray, posterior_loc: np.ndarray
    ) -> np.ndarray:
        """
        Jitted method computing degree directionality from two body-parts. E.g., ``nape`` and ``nose``,
        or ``swim_bladder`` and ``tail``.

        .. image:: _static/img/angle_from_2_bps.png
           :width: 1200
           :align: center

        :parameter np.ndarray bp_x: Size len(frames) x 2 representing x and y coordinates for first body-part.
        :parameter np.ndarray bp_y: Size len(frames) x 2 representing x and y coordinates for second body-part.
        :return np.ndarray: Frame-wise directionality in degrees.

        :example:
        >>> swim_bladder_loc = np.random.randint(low=0, high=500, size=(50, 2)).astype(np.float32)
        >>> tail_loc = np.random.randint(low=0, high=500, size=(50, 2)).astype(np.float32)
        >>> CircularStatisticsMixin().direction_two_bps(anterior_loc=swim_bladder_loc, posterior_loc=tail_loc)
        """

        results = np.full((anterior_loc.shape[0]), np.nan)
        for i in prange(anterior_loc.shape[0]):
            angle_degrees = np.degrees(
                np.arctan2(
                    anterior_loc[i][0] - posterior_loc[i][0],
                    posterior_loc[i][1] - anterior_loc[i][1],
                )
            )
            angle_degrees = angle_degrees + 360 if angle_degrees < 0 else angle_degrees
            results[i] = angle_degrees
        return results

    @staticmethod
    @njit("(float32[:],)")
    def rayleigh(data: np.ndarray) -> Tuple[float, float]:
        """
        Jitted compute of Rayleigh Z (test of non-uniformity) of single sample of circular data in degrees.

        .. note:
           Adapted from ``pingouin.circular.circ_rayleigh`` and ``pycircstat.tests.rayleigh``.

                The Rayleigh Z score is calculated as follows:

        .. math::
           Z = nR^2

        where :math:`n` is the sample size and :math:`R` is the mean resultant length.

        The associated p-value is calculated as follows:

        .. math::
           p = e^{\\sqrt{1 + 4n + 4(n^2 - R^2)} - (1 + 2n)}

        :parameter ndarray data: 1D array of size len(frames) representing degrees.
        :returns Tuple[float, float]: Tuple with Rayleigh Z score and associated probability value.

        >>> data = np.array([350, 360, 365, 360, 100, 109, 232, 123, 42, 3,4, 145]).astype(np.float32)
        >>> CircularStatisticsMixin().rayleigh(data=data)
        >>> (2.3845645695246467, 0.9842236169985417)
        """

        data = np.deg2rad(data)
        R = np.sqrt(np.sum(np.cos(data)) ** 2 + np.sum(np.sin(data)) ** 2) / len(data)
        p = np.exp(
            np.sqrt(1 + 4 * len(data) + 4 * (len(data) ** 2 - R**2))
            - (1 + 2 * len(data))
        )
        return len(data) * R**2, p

    @staticmethod
    @njit("(float32[:], float64[:], float64)")
    def rolling_rayleigh_z(
        data: np.ndarray, time_windows: np.ndarray, fps: int
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Jitted compute of Rayleigh Z (test of non-uniformity) of circular data within sliding time-window.

        .. note::
           Adapted from ``pingouin.circular.circ_rayleigh`` and ``pycircstat.tests.rayleigh``.

        :parameter ndarray data: 1D array of size len(frames) representing degrees.
        :parameter np.ndarray time_window: Rolling time-window as float in seconds. Two windows of 0.5s and 1s would be represented as np.array([0.5, 1.0])
        :parameter int fps: fps of the recorded video
        :returns Tuple[np.ndarray, np.ndarray]: Two 2d arrays with the first representing Rayleigh Z scores and second representing associated p values.

        :example:
        >>> data = np.random.randint(low=0, high=360, size=(100,)).astype(np.float32)
        >>> CircularStatisticsMixin().rolling_rayleigh_z(data=data, time_windows=np.array([0.5, 1.0]), fps=10)
        """

        data = np.deg2rad(data)
        Z_results, P_results = np.full(
            (data.shape[0], time_windows.shape[0]), 0.0
        ), np.full((data.shape[0], time_windows.shape[0]), 0.0)
        for i in prange(time_windows.shape[0]):
            win_size = int(time_windows[i] * fps)
            for j in prange(win_size, len(data) + 1):
                data_win = data[j - win_size : j]
                R = np.sqrt(
                    np.sum(np.cos(data_win)) ** 2 + np.sum(np.sin(data_win)) ** 2
                ) / len(data_win)
                Z_results[j - 1][i] = len(data_win) * R**2
                P_results[j - 1][i] = np.exp(
                    np.sqrt(1 + 4 * len(data_win) + 4 * (len(data_win) ** 2 - R**2))
                    - (1 + 2 * len(data_win))
                )
        return Z_results, P_results

    @staticmethod
    @njit("(float32[:], float32[:],)")
    def circular_correlation(sample_1: np.ndarray, sample_2: np.ndarray) -> float:
        """
        Jitted compute of circular correlation coefficient of two samples using the cross-correlation coefficient.
        Ranges from -1 to 1: 1 indicates perfect positive correlation, -1 indicates perfect negative correlation, 0
        indicates no correlation.

        .. note::
           Adapted from ``astropy.stats.circstats.circcorrcoef``.

        .. image:: _static/img/circular_correlation.png
           :width: 700
           :align: center

        :parameter np.ndarray sample_1: Angular data for e.g., Animal 1
        :parameter np.ndarray sample_1: Angular data for e.g., Animal 2
        :parameter float circular_correlation: The correlation between the two distributions.

        :example:
        >>> sample_1 = np.array([50, 90, 20, 60, 20, 90]).astype(np.float32)
        >>> sample_2 = np.array([50, 90, 70, 60, 20, 90]).astype(np.float32)
        >>> CircularStatisticsMixin().circular_correlation(sample_1=sample_1, sample_2=sample_2)
        >>> 0.7649115920066833
        """

        sample_1, sample_2 = np.deg2rad(sample_1), np.deg2rad(sample_2)
        m1 = np.arctan2(np.mean(np.sin(sample_1)), np.mean(np.cos(sample_1)))
        m2 = np.arctan2(np.mean(np.sin(sample_2)), np.mean(np.cos(sample_2)))
        sin_1, sin_2 = np.sin(sample_1 - m1), np.sin(sample_2 - m2)
        return np.sum(sin_1 * sin_2) / np.sqrt(
            np.sum(sin_1 * sin_1) * np.sum(sin_2 * sin_2)
        )

    @staticmethod
    @njit("(float32[:], float32[:], float64[:], int64)")
    def sliding_circular_correlation(
        sample_1: np.ndarray, sample_2: np.ndarray, time_windows: np.ndarray, fps: float
    ) -> np.ndarray:
        """
        Jitted compute of correlations between two angular distributions in sliding time-windows
        using the cross-correlation coefficient.

        .. image:: _static/img/cicle_correlation.png
           :width: 800
           :align: center

        :parameter np.ndarray sample_1: Angular data for e.g., Animal 1
        :parameter np.ndarray sample_1: Angular data for e.g., Animal 2
        :parameter float time_windows: Size of sliding time window in seconds. E.g., two windows of 0.5s and 1s would be represented as np.array([0.5, 1.0])
        :parameter int fps: Frame-rate of recorded video.
        :return np.ndarray: Array of size len(sample_1) x len(time_window) with correlation coefficients.


        :example:
        >>> sample_1 = np.random.randint(0, 360, (200,)).astype(np.float32)
        >>> sample_2 = np.random.randint(0, 360, (200,)).astype(np.float32)
        >>> CircularStatisticsMixin().sliding_circular_correlation(sample_1=sample_1, sample_2=sample_2, time_windows=np.array([0.5, 1.0]), fps=10.0)
        """

        sample_1, sample_2 = np.deg2rad(sample_1), np.deg2rad(sample_2)
        results = np.full((sample_1.shape[0], time_windows.shape[0]), np.nan)
        for i in prange(time_windows.shape[0]):
            win_size = int(time_windows[i] * fps)
            for j in prange(win_size, sample_1.shape[0] + 1):
                data_1_window = sample_1[j - win_size : j]
                data_2_window = sample_2[j - win_size : j]
                m1 = np.arctan2(
                    np.mean(np.sin(data_1_window)), np.mean(np.cos(data_1_window))
                )
                m2 = np.arctan2(
                    np.mean(np.sin(data_2_window)), np.mean(np.cos(data_2_window))
                )
                sin_1, sin_2 = np.sin(data_1_window - m1), np.sin(data_2_window - m2)
                c = np.sum(sin_1 * sin_2) / np.sqrt(
                    np.sum(sin_1 * sin_1) * np.sum(sin_2 * sin_2)
                )
                results[j][i] = c

        return results

    @staticmethod
    @njit("(float32[:], float64[:], int64)")
    def sliding_angular_diff(
        data: np.ndarray, time_windows: np.ndarray, fps: float
    ) -> np.ndarray:
        """
        Computes the angular difference in the current frame versus N seconds previously.
        For example, if the current angle is 45 degrees, and the angle N seconds previously was 350 degrees, then the difference
        is 55 degrees.

        .. note::
           Frames where current frame - N seconds prior equal a negative value is populated with 0.

        .. image:: _static/img/sliding_angular_difference.png
           :width: 600
           :align: center

        :parameter ndarray data: 1D array of size len(frames) representing degrees.
        :parameter np.ndarray time_window: Rolling time-window as float in seconds.
        :parameter int fps: fps of the recorded video

        :example:
        >>> data = np.array([350, 350, 1, 1]).astype(np.float32)
        >>> CircularStatisticsMixin().sliding_angular_diff(data=data, fps=1.0, time_windows=np.array([1.0]))
        >>> [[ 0.], [ 0.], [11.00000964], [ 0.]]
        """

        data_rad = np.deg2rad(data)
        results = np.full((data_rad.shape[0], time_windows.shape[0]), 0.0)
        for time_window_cnt in prange(time_windows.shape[0]):
            window_size = int(time_windows[time_window_cnt] * fps)
            for left, right in zip(
                prange(0, (data_rad.shape[0] - window_size) + 1),
                prange(window_size - 1, data_rad.shape[0] + 1),
            ):
                distance = np.pi - np.abs(
                    np.pi - np.abs(data_rad[left] - data_rad[right])
                )
                results[right][time_window_cnt] = np.rad2deg(distance)

        return results

    @staticmethod
    @njit("(float32[:], float64[:], int64)")
    def agg_angular_diff_timebins(
        data: np.ndarray, time_windows: np.ndarray, fps: int
    ) -> np.ndarray:
        """
        Compute the difference between the median angle in the current time-window versus the previous time window.
        For example, computes the difference between the mean angle in the first 1s of the video versus
        the second 1s of the video, the second 1s of the video versus the third 1s of the video, ... etc.

        .. note::
           The first time-bin of the video can't be compared against the prior time-bin and is  populated with `0`.

        .. image:: _static/img/circular_difference_time_bins.png
           :width: 800
           :align: center

        :parameter ndarray data: 1D array of size len(frames) representing degrees.
        :parameter np.ndarray time_window: Rolling time-window as float in seconds.
        :parameter int fps: fps of the recorded video

        :example:
        >>> data = np.random.normal(loc=45, scale=3, size=20).astype(np.float32)
        >>> CircularStatisticsMixin().agg_angular_diff_timebins(data=data,time_windows=np.array([1.0]), fps=5.0))
        """

        data = np.deg2rad(data)
        results = np.full((data.shape[0], time_windows.shape[0]), 0.0)
        for time_window_cnt in prange(time_windows.shape[0]):
            window_size = int(time_windows[time_window_cnt] * fps)
            prior_window = [0, window_size]
            for win_cnt, window_end in enumerate(
                prange(int(window_size * 2), data.shape[0] + 1, window_size)
            ):
                window_start = (window_end - window_size) - 1
                current_data = data[window_start:window_end]
                prior_data = data[prior_window[0] : prior_window[1]]
                prior_median = np.arctan2(
                    np.sum(np.sin(prior_data)), np.sum(np.cos(prior_data))
                )
                if prior_median < 0:
                    prior_median += 2 * np.pi
                current_median = np.arctan2(
                    np.sum(np.sin(current_data)), np.sum(np.cos(current_data))
                )
                if current_median < 0:
                    current_median += 2 * np.pi
                distance = np.pi - np.abs(np.pi - np.abs(prior_median - current_median))
                results[window_start:window_end, win_cnt] = np.rad2deg(distance)
                prior_window = [window_start, window_end]

        return results

    @staticmethod
    @njit("(float32[:], float64[:], int64)")
    def sliding_rao_spacing(
        data: np.ndarray, time_windows: np.ndarray, fps: int
    ) -> np.ndarray:
        """
        Jitted compute of the uniformity of a circular dataset in sliding windows.

        :parameter ndarray data: 1D array of size len(frames) representing degrees.
        :parameter np.ndarray time_window: Rolling time-window as float in seconds.
        :parameter int fps: fps of the recorded video
        :return np.ndarray: representing rao-spacing U in every sliding windows [-window:n]

        .. image:: _static/img/raospacing.png
           :width: 800
           :align: center

        The Rao's Spacing (:math:`U`) is calculated as follows:

        .. math::

           U = \\frac{1}{2} \\sum_{i=1}^{N} |l - T_i|

        where :math:`N` is the number of data points in the sliding window, :math:`T_i` is the spacing between adjacent data points, and :math:`l` is the equal angular spacing.

        :references:
        .. [1] `UCSB <https://jammalam.faculty.pstat.ucsb.edu/html/favorite/test.htm>`__.

        :example:
        >>> data = np.random.randint(low=0, high=360, size=(500,)).astype(np.float32)
        >>> result = CircularStatisticsMixin().sliding_rao_spacing(data=data, time_windows=np.array([0.5, 1.0]), fps=10)
        """

        results = np.full((data.shape[0], time_windows.shape[0]), -1.0)
        for win_cnt in prange(time_windows.shape[0]):
            window_size = int(time_windows[win_cnt] * fps)
            for i in range(window_size, data.shape[0]):
                w_data = np.sort(data[i - window_size : i])
                Ti, TiL = np.full((w_data.shape[0]), np.nan), np.full(
                    (w_data.shape[0]), np.nan
                )
                l = np.int8(360 / len(w_data))
                Ti[-1] = np.rad2deg(
                    np.pi
                    - np.abs(
                        np.pi - np.abs(np.deg2rad(w_data[0]) - np.deg2rad(w_data[-1]))
                    )
                )
                for j in prange(w_data.shape[0] - 1, -1, -1):
                    Ti[j] = np.rad2deg(
                        np.pi
                        - np.abs(
                            np.pi
                            - np.abs(np.deg2rad(w_data[j]) - np.deg2rad(w_data[j - 1]))
                        )
                    )
                for k in prange(Ti.shape[0]):
                    TiL[int(k)] = max((l, Ti[k])) - min((l, Ti[k]))
                S = np.sum(TiL)
                U = int(S / 2)
                results[i][win_cnt] = U
        return results

    @staticmethod
    @njit("(float32[:], float32[:])")
    def kuipers_two_sample_test(sample_1: np.ndarray, sample_2: np.ndarray) -> float:
        """
        Compute the Kuiper's two-sample test statistic for circular distributions.

        Kuiper's two-sample test is a non-parametric test used to determine if two samples are drawn from the same circular distribution. It is particularly useful for circular data, such as angles or directions.

        .. note::
           Adapted from `Kuiper <https://github.com/aarchiba/kuiper/tree/master>`__ by `Anne Archibald <https://github.com/aarchiba>`_.

        :parameter ndarray data: The first circular sample array in degrees.
        :parameter ndarray data: The second circular sample array in degrees.
        :return float: Kuiper's test statistic.

        :example:
        >>> sample_1, sample_2 = np.random.normal(loc=45, scale=1, size=100).astype(np.float32), np.random.normal(loc=180, scale=20, size=100).astype(np.float32)
        >>> CircularStatisticsMixin().kuipers_two_sample_test(sample_1=sample_1, sample_2=sample_2)
        """

        sample_1, sample_2 = np.deg2rad(np.sort(sample_1)), np.deg2rad(
            np.sort(sample_2)
        )
        cdfv1 = np.searchsorted(sample_2, sample_1) / float(len(sample_2))
        cdfv2 = np.searchsorted(sample_1, sample_2) / float(len(sample_1))
        return np.amax(
            cdfv1 - np.arange(len(sample_1)) / float(len(sample_1))
        ) + np.amax(cdfv2 - np.arange(len(sample_2)) / float(len(sample_2)))

    @staticmethod
    @njit("(float32[:], float32[:], float64[:], int64)")
    def sliding_kuipers_two_sample_test(
        sample_1: np.ndarray, sample_2: np.ndarray, time_windows: np.ndarray, fps: int
    ) -> np.ndarray:
        """
        Jitted compute of Kuipers two-sample test comparing two distributions with sliding time window.

        This function calculates the Kuipers two-sample test statistic for each time window, sliding through the given circular data sequences.

        :parameter np.ndarray data: The first circular sample array in degrees.
        :parameter np.ndarray data: The second circular sample array in degrees.
        :parameter np.ndarray time_windows: An array containing the time window sizes (in seconds) for which the Kuipers two-sample test will be computed.
        :parameter int fps: The frames per second, representing the sampling rate of the data.
        :returns np.ndarray: A 2D array containing the Kuipers two-sample test statistics for each time window and each time step.

        :examples:
        >>> data = np.random.randint(low=0, high=360, size=(100,)).astype(np.float64)
        >>> D = CircularStatisticsMixin().sliding_kuipers_two_sample_test(data=data, time_windows=np.array([0.5, 5]), fps=2)
        """
        sample_1, sample_2 = np.deg2rad(sample_1), np.deg2rad(sample_2)
        results = np.full((sample_1.shape[0], time_windows.shape[0]), -1.0)
        for time_window_cnt in prange(time_windows.shape[0]):
            win_size = int(time_windows[time_window_cnt] * fps)
            for i in range(win_size, sample_1.shape[0]):
                sample_1_win, sample_2_win = (
                    sample_1[i - win_size : i],
                    sample_2[i - win_size : i],
                )
                cdfv1 = np.searchsorted(sample_2, sample_1_win) / float(
                    len(sample_2_win)
                )
                cdfv2 = np.searchsorted(sample_1_win, sample_2_win) / float(
                    len(sample_1_win)
                )
                D = np.amax(
                    cdfv1 - np.arange(len(sample_1_win)) / float(len(sample_1_win))
                ) + np.amax(
                    cdfv2 - np.arange(len(sample_2_win)) / float(len(sample_2_win))
                )
                results[i][time_window_cnt] = D

        return results

    @staticmethod
    def sliding_hodges_ajne(
        data: np.ndarray, time_window: float, fps: int
    ) -> np.ndarray:

        data = np.deg2rad(data)
        results, window_size = np.full((data.shape[0]), -1.0), int(time_window * fps)
        for i in range(window_size, data.shape[0]):
            w_data = data[i - window_size : i]
            v = 1 - np.abs(np.mean(np.exp(1j * w_data)))
            n = len(w_data)
            H = n * (1 - v)
            results[i] = H
        return results

    @staticmethod
    def hodges_ajne(sample: np.ndarray):
        v = 1 - np.abs(np.mean(np.exp(1j * sample)))
        n = len(sample)
        H = n * (1 - v)
        return H

    @staticmethod
    @jit(nopython=True)
    def watson_williams_test(sample_1: np.ndarray, sample_2: np.ndarray):

        variance1 = 1 - np.abs(np.mean(np.exp(1j * sample_1)))
        variance2 = 1 - np.abs(np.mean(np.exp(1j * sample_2)))
        numerator = (variance1 + variance2) / 2
        denominator = (variance1**2 / len(sample_1)) + (variance2**2 / len(sample_2))
        F = numerator / denominator
        return F

    @staticmethod
    def watsons_u(data: np.ndarray):
        data = np.deg2rad(data)
        mean_vector = np.exp(1j * data).mean()
        n = len(data)
        return n * (1 - np.abs(mean_vector))

    @staticmethod
    @njit("(float32[:],)")
    def circular_range(data: np.ndarray) -> float:
        """
        Jitted compute of circular range in degrees. The range is defined as the angular span of the
        shortest arc that can contain all the data points. A smaller range indicates a more concentrated
        distribution, while a larger range suggests a more dispersed distribution.

        .. image:: _static/img/circular_range.png
           :width: 600
           :align: center

        :parameter ndarray data: 1D array of circular data measured in degrees
        :return np.ndarray: The circular range in degrees.

        :example:
        >>> CircularStatisticsMixin().circular_range([350, 20, 60, 100])
        >>> 110.0
        >>> CircularStatisticsMixin().circular_range([110, 20, 60, 100])
        >>> 90.0
        """

        data = np.sort(np.deg2rad(data))
        angular_diffs = np.diff(data)
        circular_range = np.max(angular_diffs)
        return np.ceil(np.rad2deg(min(2 * np.pi - circular_range, data[-1] - data[0])))

    @staticmethod
    @njit("(float32[:], float64[:], int64)")
    def sliding_circular_range(
        data: np.ndarray, time_windows: np.ndarray, fps: int
    ) -> np.ndarray:
        """
        Jitted compute of sliding circular range for a time series of circular data. The range is defined as the angular span of the
        shortest arc that can contain all the data points. Measures the circular spread of data within sliding time windows of specified duration.

        .. image:: _static/img/sliding_circular_range.png
           :width: 600
           :align: center

        :parameter np.ndarray data: 1D array of circular data measured in degrees
        :parameter np.ndarray time_windows: Size of sliding time window in seconds. E.g., two windows of 0.5s and 1s would be represented as np.array([0.5, 1.0])
        :parameter int fps: Frame-rate of recorded video.
        :return np.ndarray: Array of size len(sample_1) x len(time_window) with angular ranges in degrees.

        :examples:
        >>> data = np.array([260, 280, 300, 340, 360, 0, 10, 350, 0, 15]).astype(np.float32)
        >>> CircularStatisticsMixin().sliding_circular_range(data=data, time_windows=np.array([0.5]), fps=10)
        >>> [[ -1.],[ -1.],[ -1.],[ -1.],[100.],[80],[70],[30],[20],[25]]
        """

        results = np.full((data.shape[0], time_windows.shape[0]), -1.0)
        for time_window_cnt in prange(time_windows.shape[0]):
            win_size = int(time_windows[time_window_cnt] * fps)
            for left, right in zip(
                prange(0, (data.shape[0] - win_size) + 1),
                prange(win_size - 1, data.shape[0] + 1),
            ):
                sample = np.sort(np.deg2rad(data[left : right + 1]))
                angular_diffs = np.diff(sample)
                circular_range = np.max(angular_diffs)
                results[right][time_window_cnt] = np.rad2deg(
                    min(2 * np.pi - circular_range, sample[-1] - sample[0])
                )
        return results

    @staticmethod
    @njit("(float32[:], int64[:, :],)")
    def circular_hotspots(data: np.ndarray, bins: np.ndarray) -> np.ndarray:
        """
        Calculate the proportion of data points falling within circular bins.

        :parameter ndarray data: 1D array of circular data measured in degrees.
        :parameter ndarray bins: 2D array of shape representing circular bins defining [start_degree, end_degree] inclusive..
        :return np.ndarray: 1D array containing the proportion of data points that fall within each specified circular bin.

        .. image:: _static/img/circular_hotspots.png
           :width: 700
           :align: center

        :example:
        >>> data = np.array([270, 360, 10, 90, 91, 180, 185, 260]).astype(np.float32)
        >>> bins = np.array([[270, 90], [91, 268]])
        >>> CircularStatisticsMixin().circular_hotspots(data=data, bins=bins)
        >>> [0.5, 0.5]
        >>> bins = np.array([[270, 0], [1, 90], [91, 180], [181, 269]])
        >>> CircularStatisticsMixin().circular_hotspots(data=data, bins=bins)
        >>> [0.25, 0.25, 0.25, 0.25]
        """
        results = np.full((bins.shape[0]), -1.0)
        for bin_cnt in range(bins.shape[0]):
            if bins[bin_cnt][0] > bins[bin_cnt][1]:
                mask = ((data >= bins[bin_cnt][0]) & (data <= 360)) | (
                    (data >= 0) & (data <= bins[bin_cnt][1])
                )
            else:
                mask = (data >= bins[bin_cnt][0]) & (data <= bins[bin_cnt][1])
            results[bin_cnt] = data[mask].shape[0] / data.shape[0]
        return results

    @staticmethod
    @njit("(float32[:], int64[:, :], float64, float64)")
    def sliding_circular_hotspots(
        data: np.ndarray, bins: np.ndarray, time_window: float, fps: float
    ) -> np.ndarray:
        """
        Jitted compute of sliding circular hotspots in a dataset. Calculates circular hotspots in a time-series dataset by sliding a time window
        across the data and computing hotspot statistics for specified circular bins.

        :parameter ndarray data: 1D array of circular data measured in degrees.
        :parameter ndarray bins: 2D array of shape representing circular bins defining [start_degree, end_degree] inclusive.
        :parameter float time_window: The size of the sliding window in seconds.
        :parameter float fps: The frame-rate of the video.

        :return np.ndarray: A 2D numpy array where each row corresponds to a time point in `data`, and each column represents a circular bin. The values in the array represent the proportion of data points
                            within each bin at each time point. If a bin contains no data points in the window, the
                            corresponding value will be -1.0.

        .. note::
           - The function utilizes the Numba JIT compiler for improved performance.
           - Circular bin definitions should follow the convention where angles are specified in degrees
             within the range [0, 360], and the bins are defined using start and end angles inclusive.
             For example, (0, 90) represents the first quadrant in a circular space.


        .. image:: _static/img/sliding_circular_hotspot.png
           :width: 600
           :align: center

        :example:
        >>> data = np.array([270, 360, 10, 20, 90, 91, 180, 185, 260, 265]).astype(np.float32)
        >>> bins = np.array([[270, 90], [91, 268]])
        >>> CircularStatisticsMixin().sliding_circular_hotspots(data=data, bins=bins, time_window=0.5, fps=10)
        >>> [[-1. , -1. ],
        >>>  [-1. , -1. ],
        >>>  [-1. , -1. ],
        >>>  [-1. , -1. ],
        >>>  [ 0.5,  0. ],
        >>>  [ 0.4,  0.1],
        >>>  [ 0.3,  0.2],
        >>>  [ 0.2,  0.3],
        >>>  [ 0.1,  0.4],
        >>>  [ 0. ,  0.5]]

        """
        results = np.full((data.shape[0], bins.shape[0]), -1.0)
        win_size = int(time_window * fps)
        for left, right in zip(
            prange(0, (data.shape[0] - win_size) + 1),
            prange(win_size - 1, data.shape[0] + 1),
        ):
            sample = data[left : right + 1]
            for bin_cnt in range(bins.shape[0]):
                if bins[bin_cnt][0] > bins[bin_cnt][1]:
                    mask = ((sample >= bins[bin_cnt][0]) & (sample <= 360)) | (
                        (sample >= 0) & (sample <= bins[bin_cnt][1])
                    )
                else:
                    mask = (sample >= bins[bin_cnt][0]) & (sample <= bins[bin_cnt][1])
                results[right][bin_cnt] = data[mask].shape[0] / data.shape[0]
        return results

    @staticmethod
    @njit("(float32[:],)")
    def rotational_direction(data: np.ndarray) -> np.ndarray:
        """
        Jitted compute of frame-by-frame rotational direction within a 1D timeseries array of angular data.

        :parameter ndarray data: 1D array of size len(frames) representing degrees.
        :return numpy.ndarray: An array of directional indicators.
           - 0 indicates no rotational change relative to prior frame.
           - 1 indicates a clockwise rotational change relative to prior frame.
           - 2 indicates a counter-clockwise rotational change relative to prior frame.

        .. note::
           * For the first frame, no rotation is possible so is populated with -1.
           * Frame-by-frame rotations of 180° degrees are denoted as clockwise rotations.

        .. image:: _static/img/rotational_direction.png
           :width: 600
           :align: center

        :example:
        >>> data = np.array([45, 50, 35, 50, 80, 350, 350, 0 , 180]).astype(np.float32)
        >>> CircularStatisticsMixin().rotational_direction(data)
        >>> [-1.,  1.,  2.,  1.,  1.,  2.,  0.,  1.,  1.]
        """

        data = data % 360
        data_rad = np.deg2rad(data)
        result, prior_angle = np.full((data.shape[0]), -1.0), data_rad[0]
        for i in prange(1, data.shape[0]):
            current_angle = data_rad[i]
            distance = np.round(
                np.rad2deg(np.pi - np.abs(np.pi - np.abs(current_angle - prior_angle)))
            )
            if data[i - 1] == data[i]:
                result[i] = 0
            elif data[i - 1] == (data[i] - distance) % 360:
                result[i] = 1
            else:
                result[i] = 2
            prior_angle = current_angle
        return result

    @staticmethod
    @njit("(int64[:, :, :], int64)")
    def fit_circle(data: np.ndarray, max_iterations: int) -> np.ndarray:
        """
        Fit a circle to a dataset using the least squares method.

        This function fits a circle to a dataset using the least squares method. The circle is defined
        by the equation:

        .. math::
           X^2 + Y^2 = R^2

        .. note::
           Adapted to numba JIT from `circle-fit <https://github.com/AlliedToasters/circle-fit>`_ ``hyperLSQ`` method.

        .. image:: _static/img/fit_circle.png
           :width: 600
           :align: center

        References
        ----------

        .. [1] Kanatani, Rangarajan, Hyper least squares fitting of circles and ellipses, `Computational Statistics & Data Analysis`,
               vol. 55, pp. 2197-2208, 2011.
        .. [2] Lapp, Salazar, Champagne. Automated maternal behavior during early life in rodents (AMBER) pipeline, `Scientific Reports`,
               13:18277, 2023.

        :parameter np.ndarray data: A 3D NumPy array with shape (N, M, 2). N represent frames, M represents the number of body-parts, and 2 represents x and y coordinates.
        :parameter int max_iterations: The maximum number of iterations for fitting the circle.
        :returns np.ndarray: Array with shape (N, 3) with N representing frame and 3 representing (i) X-coordinate of the circle center, (ii) Y-coordinate of the circle center, and (iii) Radius of the circle

        :example:
        >>> data = np.array([[[5, 10], [10, 5], [15, 10], [10, 15]]])
        >>> CircularStatisticsMixin().fit_circle(data=data, iter_max=88)
        >>> [[10, 10, 5]]
        """

        results = np.full((data.shape[0], 3), np.nan)
        for i in range(data.shape[0]):
            frm_data = data[i]
            x, y, n = frm_data[:, 0], frm_data[:, 1], frm_data.shape[0]

            Xi, Yi = x - x.mean(), y - y.mean()
            Zi = Xi * Xi + Yi * Yi
            Mxx, Mxy, Mxz = (
                (Xi * Xi).sum() / n,
                (Xi * Yi).sum() / n,
                (Xi * Zi).sum() / n,
            )
            Myy, Myz = (Yi * Yi).sum() / n, (Yi * Zi).sum() / n
            Mzz = (Zi * Zi).sum() / n

            Mz = Mxx + Myy
            Var_z = Mzz - Mz * Mz
            Cov_xy = Mxx * Myy - Mxy * Mxy

            A2 = 4 * Cov_xy - 3 * Mz * Mz - Mzz
            A1 = Var_z * Mz + 4.0 * Cov_xy * Mz - Mxz * Mxz - Myz * Myz
            A0 = (
                Mxz * (Mxz * Myy - Myz * Mxy)
                + Myz * (Myz * Mxx - Mxz * Mxy)
                - Var_z * Cov_xy
            )
            A22 = A2 + A2

            Y, X = A0, 0.0
            for j in range(max_iterations):
                Dy = A1 + X * (A22 + 16.0 * (X**2))
                xnew = X - Y / Dy
                if xnew == X or not np.isfinite(xnew):
                    break
                ynew = A0 + xnew * (A1 + xnew * (A2 + 4.0 * xnew * xnew))
                if abs(ynew) >= abs(Y):
                    break
                X, Y = xnew, ynew

            det = X**2 - X * Mz + Cov_xy
            Xcenter = (Mxz * (Myy - X) - Myz * Mxy) / det / 2.0
            Ycenter = (Myz * (Mxx - X) - Mxz * Mxy) / det / 2.0

            results[i, :] = np.array(
                [
                    Xcenter + x.mean(),
                    Ycenter + y.mean(),
                    np.sqrt(abs(Xcenter**2 + Ycenter**2 + Mz)),
                ]
            )

        return results.astype(np.int64)


# data_sizes = [1, 10, 100, 1000, 10000, 100000, 1000000, 10000000, 100000000]
# # #
# # for data_size in data_sizes:
# #     data = np.random.randint(low=0, high=360, size=(data_size,)).astype(np.float32)
# #     start = time.time()
# #     CircularStatisticsMixin().rayleigh(data=data)
# #     print(time.time() - start)
#
#
# for data_size in data_sizes:
#     data = np.random.randint(low=0, high=360, size=(data_size,)).astype(np.float32)
#     start = time.time()
#     CircularStatisticsMixin().rolling_rayleigh_z(data=data, fps=10, time_windows=np.array([0.5, 1.0]))
#     print(time.time() - start)


# for data_size in data_sizes:
#     data = np.random.randint(low=0, high=360, size=(data_size,)).astype(np.float32)
#     start = time.time()
#     CircularStatisticsMixin().degrees_to_cardinal(degree_angles=data)
#     print(time.time() - start)


# for data_size in data_sizes:
#     nose_loc = np.random.randint(low=0, high=500, size=(data_size, 2)).astype(np.float32)
#     left_ear_loc = np.random.randint(low=0, high=500, size=(data_size, 2)).astype(np.float32)
#     right_ear_loc = np.random.randint(low=0, high=500, size=(data_size, 2)).astype(np.float32)
#     start = time.time()
#     results = CircularStatisticsMixin().direction_two_bps(swim_bladder_loc=nose_loc,tail_loc=left_ear_loc)
#     print(time.time() - start)



import os
import platform
from configparser import ConfigParser
from tkinter import *

from simba.roi_tools import ROI_image
from simba.roi_tools.ROI_define import ROI_definitions
from simba.roi_tools.ROI_image import ROI_image_class
from simba.roi_tools.ROI_multiply import multiply_ROIs
from simba.roi_tools.ROI_reset import reset_video_ROIs
from simba.ui.tkinter_functions import CreateLabelFrameWithIcon
from simba.utils.enums import ConfigKey, Keys, Links
from simba.utils.errors import NoFilesFoundError
from simba.utils.lookups import get_color_dict


class ROI_menu:
    def __init__(self, config_path, new_roi=True):
        self.config_path = config_path
        config = ConfigParser()
        config.read(config_path)
        self.project_path = config.get(
            ConfigKey.GENERAL_SETTINGS.value, ConfigKey.PROJECT_PATH.value
        )
        self.measures_dir = os.path.join(self.project_path, "logs", "measures")
        self.video_dir = os.path.join(self.project_path, "videos")
        self.duplicate_jump_size = 20
        self.click_sens = 10
        self.text_size = 5
        self.text_thickness = 3
        self.line_type = -1
        self.named_shape_colors = get_color_dict()
        self.roi_table_menu()

    def roi_table_menu(self):
        self.filesFound = []
        self.row = []
        for i in os.listdir(self.video_dir):
            if i.endswith((".avi", ".mp4", ".mov", "flv")):
                self.filesFound.append(i)
        if len(self.filesFound) == 0:
            raise NoFilesFoundError(
                "No videos found the SimBA project (no avi, mp4, mov, flv files in the project_folder/videos directory)"
            )

        maxname = max(self.filesFound, key=len)
        self.roimenu = Toplevel()
        self.roimenu.minsize(720, 960)
        self.roimenu.wm_title("ROI Table")

        self.scroll_window = hxtScrollbar(self.roimenu)
        tableframe = CreateLabelFrameWithIcon(
            parent=self.scroll_window,
            header="Video Name",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.ROI.value,
        )
        # tableframe = LabelFrame(self.scroll_window, text='Video Name', labelanchor=NW)

        self.image_data = ROI_image_class(
            self.config_path,
            self.named_shape_colors,
            self.duplicate_jump_size,
            self.line_type,
            self.click_sens,
            self.text_size,
            self.text_thickness,
        )
        for i in range(len(self.filesFound)):
            self.row.append(
                roitableRow(
                    self.image_data,
                    tableframe,
                    self.video_dir,
                    str(self.filesFound[i]),
                    str(len(maxname) + 20),
                    str(i + 1) + ".",
                    projectini=self.config_path,
                )
            )
            self.row[i].grid(row=i + 1, sticky=W)
        tableframe.grid(row=0)


class roitableRow(Frame):
    def __init__(
        self,
        image_data: ROI_image,
        parent=None,
        dirname="",
        filename="",
        widths="",
        indexs="",
        projectini="",
    ):
        self.image_data = image_data
        self.projectini = projectini
        self.filename = os.path.join(dirname, filename)
        Frame.__init__(self, master=parent)
        if not os.path.exists(self.filename):
            if os.path.islink(self.filename):
                filename += " symlink"
            filename += " path not exists"
        var = StringVar()
        self.index = Entry(self, textvariable=var, width=4)
        var.set(indexs)
        self.index.grid(row=0, column=0)
        self.lblName = Label(self, text=filename, width=widths, anchor=W)
        self.lblName.grid(row=0, column=1, sticky=W)
        if os.path.exists(self.filename):
            self.btnset = Button(self, text="Draw", command=self.draw)
            self.btnset.grid(row=0, column=2)
            self.btnreset = Button(self, text="Reset", command=self.reset)
            self.btnreset.grid(row=0, column=3)
            self.btnapplyall = Button(self, text="Apply to all", command=self.applyall)
            self.btnapplyall.grid(row=0, column=4)

    def draw(self):
        ROI_definitions(self.projectini, self.filename, self.image_data)

    def reset(self):
        reset_video_ROIs(self.projectini, self.filename)

    def applyall(self):
        multiply_ROIs(self.projectini, self.filename)


def onMousewheel(event, canvas):
    try:
        scrollSpeed = event.delta
        if platform.system() == "Darwin":
            scrollSpeed = event.delta
        elif platform.system() == "Windows":
            scrollSpeed = int(event.delta / 120)
        canvas.yview_scroll(-1 * (scrollSpeed), "units")
    except:
        pass


def bindToMousewheel(event, canvas):
    canvas.bind_all("<MouseWheel>", lambda event: onMousewheel(event, canvas))


def unbindToMousewheel(event, canvas):
    canvas.unbind_all("<MouseWheel>")


def onFrameConfigure(canvas):
    """Reset the scroll region to encompass the inner frame"""
    canvas.configure(scrollregion=canvas.bbox("all"))


def hxtScrollbar(master):
    """
    Create canvas.
    Create a frame and put it in the canvas.
    Create two scrollbar and insert command of canvas x and y view
    Use canvas to create a window, where window = frame
    Bind the frame to the canvas
    """

    bg = master.cget("background")
    acanvas = Canvas(master, borderwidth=0, background=bg)
    vsb = Scrollbar(master, orient="vertical", command=acanvas.yview)
    vsb2 = Scrollbar(master, orient="horizontal", command=acanvas.xview)
    acanvas.configure(yscrollcommand=vsb.set)
    acanvas.configure(xscrollcommand=vsb2.set)
    vsb.pack(side=RIGHT, fill="y")
    vsb2.pack(side=BOTTOM, fill="x")
    acanvas.pack(side="left", fill="both", expand=True)

    # bind the frame to the canvas
    acanvas.bind("<Configure>", lambda event, canvas=acanvas: onFrameConfigure(acanvas))
    acanvas.bind("<Enter>", lambda event: bindToMousewheel(event, acanvas))
    acanvas.bind("<Leave>", lambda event: unbindToMousewheel(event, acanvas))
    frame = Frame(acanvas, background=bg)
    acanvas.create_window((10, 10), window=frame, anchor="nw")

    return frame



__author__ = "Simon Nilsson"

from tkinter import *

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.roi_tools.ROI_size_standardizer import ROISizeStandardizer
from simba.ui.tkinter_functions import CreateLabelFrameWithIcon, DropDownMenu
from simba.utils.checks import check_file_exist_and_readable
from simba.utils.enums import Keys, Links
from simba.utils.errors import NoROIDataError


class ROISizeStandardizerPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str):
        PopUpMixin.__init__(self, title="ROI SIZE NORMALIZER PY PIXELS PER MILLIMETER")
        ConfigReader.__init__(self, config_path=config_path)
        check_file_exist_and_readable(file_path=self.roi_coordinates_path)
        self.read_roi_data()
        if len(self.video_names_w_rois) <= 1:
            raise NoROIDataError(
                f"SimBA could only find {len(self.video_names_w_rois)} videos with ROIs in your project. You need at least 2 videos with ROIs to standardize ROI sizes"
            )
        self.video_names = list(self.video_names_w_rois)
        self.settings_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="SETTINGS",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.ROI_FEATURES_PLOT.value,
        )
        self.baseline_video_dropdown = DropDownMenu(
            self.settings_frm, "Baseline video:", self.video_names, "12"
        )
        self.baseline_video_dropdown.setChoices(self.video_names[0])

        self.settings_frm.grid(row=0, column=0, sticky=NW)
        self.baseline_video_dropdown.grid(row=0, column=0, sticky=NW)

        self.create_run_frm(run_function=self.run)
        # self.main_frm.mainloop()

    def run(self):
        roi_standardizer = ROISizeStandardizer(
            config_path=self.config_path,
            reference_video=self.baseline_video_dropdown.getChoices(),
        )
        roi_standardizer.run()
        roi_standardizer.save()


# ROISizeStandardizerPopUp(config_path="/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini")



import os
from typing import Union

import pandas as pd

from simba.utils.checks import check_if_dir_exists, check_that_column_exist
from simba.utils.printing import stdout_success, stdout_warning
from simba.utils.read_write import (find_files_of_filetypes_in_directory,
                                    get_fn_ext, write_df)

"""
INSTRUCTIONS:
1. Change the DATA_DIR to the directory where you have your BORIS annotation files on line 8.
2. Change the SAVE_DIR to the directory where you want to save your new fixed BORIS files on line 9
3. If the names of the subjects or behavior columns changes, change these two names to the names of the columns.
4. The SETTINGS specifies to example two rules, (i) for every row where the subject is focal and the behavior is invest_no, change the behavior name to focal_invest_no, and 
   (ii) for every row where the subject is stimulus and the behavior is invest_no, change the behavior name to stimulus_invest_no. Change these rules or add more if you have more behaviors or animals.
5. Save the file.
6. Activate the SimBA conda environment and navigate to the directory where the boris_source_cleaner.py is located.
7. Run it by typing python boris_source_cleaner.py
"""

DATA_DIR = r"/Users/simon/Downloads/boris_data"  # Directory with BORIS annotations in CSV format.
SAVE_DIR = r"/Users/simon/Downloads/save_dir"  # Directory where to save the modified BORIS annotations in CSV format.
SUBJECT_COL = "Subject"  # The name of the Subject column in BORIS files. Change this if the animal names are encoded in a different column than *Subject*.
BEHAVIOR_COL = "Behavior"  # The name of the Behavior column in BORIS files. Change this if the behavior names are encoded in a different column than *Behavior*.

SETTINGS = [
    {
        SUBJECT_COL: "focal",
        BEHAVIOR_COL: "invest_no",
        "New_behavior_name": "focal_invest_no",
    },
    {
        SUBJECT_COL: "stimulus",
        BEHAVIOR_COL: "invest_no",
        "New_behavior_name": "stimulus_invest_no",
    },
]

# SUBJECT_COL: The entry in the *Subject* column. For example, animal "Simon" or "stimulus"
# BEHAVIOR_COL: The entry in the *Behavior* column. For example, "behavior1" or "invest_no"
# New_behavior_name: For rows where the subjects column and behavior column critera filled, change the behavior column this new behavior name.
# E.g., for rows where *Subject* is Simon AND *Behavior* is behavior1, the *Behavior* will be changed to Simon_behavior1.


class BorisSourceCleaner(object):
    """
    Helper to clean BORIS files where behavior with the same name are assigned to different
    animals through the **Subjects** field.

    :parameter Union[str, os.PathLike] data_dir: Directory with BORIS annotations in CSV format.
    :parameter Union[str, os.PathLike] save_dir: Directory to store the modified BORIS annotations in CSV format.
    :parameter dict settings: Rules for how to change the behavior names.

    :example:
    >>> boris_cleaner = BorisSourceCleaner(data_dir='/Users/simon/Downloads/boris_data', save_dir='/Users/simon/Downloads/save_dir', settings=SETTINGS)
    >>> boris_cleaner.run()

    """

    def __init__(
        self,
        data_dir: Union[str, os.PathLike],
        save_dir: Union[str, os.PathLike],
        settings: dict,
    ):
        check_if_dir_exists(in_dir=data_dir)
        check_if_dir_exists(in_dir=save_dir)
        self.data_paths = find_files_of_filetypes_in_directory(
            directory=data_dir,
            extensions=[".csv"],
            raise_warning=False,
            raise_error=True,
        )
        self.settings, self.save_dir = settings, save_dir

    def run(self):
        for file_cnt, file_path in enumerate(self.data_paths):
            video_name = get_fn_ext(filepath=file_path)[1]
            save_path = os.path.join(self.save_dir, f"{video_name}.csv")
            df = pd.read_csv(file_path)
            check_that_column_exist(df=df, column_name=SUBJECT_COL, file_name=file_path)
            check_that_column_exist(
                df=df, column_name=BEHAVIOR_COL, file_name=file_path
            )

            for rule in self.settings:
                rule_rows = df.loc[
                    (df[SUBJECT_COL] == rule[SUBJECT_COL])
                    & (df[BEHAVIOR_COL] == rule[BEHAVIOR_COL])
                ]
                rule_rows = rule_rows[rule_rows["Behavior type"] != "POINT"]

                if len(rule_rows) == 0:
                    stdout_warning(
                        msg=f"WARNING: No rows for subject name {rule[SUBJECT_COL]} with behavior name {rule[BEHAVIOR_COL]} found in file {file_path}"
                    )
                    continue

                elif (len(rule_rows) % 2) != 0:
                    stdout_warning(
                        msg=f"WARNING: Found an UNEVEN number of rows for subject name {rule[SUBJECT_COL]} and behavior name {rule[BEHAVIOR_COL]} found in file {file_path}"
                    )

                rule_rows[BEHAVIOR_COL] = rule["New_behavior_name"]
                df.update(rule_rows)

            write_df(df=df, file_type="csv", save_path=save_path)
            stdout_success(
                msg=f"SAVED DATA VIDEO {save_path}. Video {file_cnt+1}/{len(self.data_paths)}"
            )

        stdout_success(f"All data files saved in {self.save_dir}.")


if __name__ == "__main__":
    boris_cleaner = BorisSourceCleaner(
        data_dir=DATA_DIR, save_dir=SAVE_DIR, settings=SETTINGS
    )
    boris_cleaner.run()



__author__ = "Simon Nilsson"

import warnings

warnings.filterwarnings("ignore", category=FutureWarning)
warnings.filterwarnings("ignore", category=DeprecationWarning)
import functools
import multiprocessing
import os
import platform
from typing import Any, Dict

import numpy as np

from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.mixins.train_model_mixin import TrainModelMixin
from simba.utils.data import plug_holes_shortest_bout
from simba.utils.printing import stdout_success
from simba.utils.read_write import (concatenate_videos_in_folder, get_fn_ext,
                                    get_video_meta_data, read_df, write_df)


class ValidateModelOneVideoMultiprocess(ConfigReader, PlottingMixin, TrainModelMixin):
    """
    Create classifier validation video for a single input video. Results are stored in the
    ``project_folder/frames/output/validation`` directory.

    :param str config_path: path to SimBA project config file in Configparser format
    :param str feature_file_path: path to SimBA file (parquet or CSV) containing pose-estimation and feature fields.
    :param str model_path: path to pickled classifier object
    :param float discrimination_threshold: classification threshold.
    :param int shortest_bout: Allowed classified bout length expressed in milliseconds. E.g., `1000` will shift frames classified
        as containing the behavior, but occuring in a bout shorter than `1000`, from `target present to `target absent`.
    :param str create_gantt:
        If SimBA should create gantt charts alongside the validation video. OPTIONS: 'None', 'Gantt chart: final frame only (slightly faster)',
        'Gantt chart: video'.
    :param dict settings: User style settings for video. E.g., {'pose': True, 'animal_names': True, 'styles': None}
    :param int cores: Number of cores to use.

    :example:
    >>> test = ValidateModelOneVideoMultiprocess(config_path=r'/Users/simon/Desktop/envs/troubleshooting/naresh/project_folder/project_config.ini',
    >>>                                         feature_file_path=r'/Users/simon/Desktop/envs/troubleshooting/naresh/project_folder/csv/features_extracted/SF2.csv',
    >>>                                         model_path='/Users/simon/Desktop/envs/troubleshooting/naresh/models/generated_models/Top.sav',
    >>>                                         discrimination_threshold=0.6,
    >>>                                         shortest_bout=50,
    >>>                                         settings={'pose': True, 'animal_names': True, 'styles': None},
    >>>                                         create_gantt='Gantt chart: final frame only (slightly faster)',
    >>>                                         cores=5)
    >>> test.run()
    """

    def __init__(
        self,
        config_path: str,
        feature_file_path: str,
        model_path: str,
        discrimination_threshold: float,
        shortest_bout: int,
        cores: int,
        create_gantt: str,
        settings: Dict[str, Any],
    ):
        ConfigReader.__init__(self, config_path=config_path)
        PlottingMixin.__init__(self)
        TrainModelMixin.__init__(self)
        if platform.system() == "Darwin":
            multiprocessing.set_start_method("spawn", force=True)
        _, self.feature_filename, ext = get_fn_ext(feature_file_path)
        (
            self.discrimination_threshold,
            self.cores,
            self.shortest_bout,
            self.create_gantt,
            self.settings,
        ) = (
            float(discrimination_threshold),
            cores,
            shortest_bout,
            create_gantt,
            settings,
        )
        if self.create_gantt == "None":
            self.create_gantt = None
        if not os.path.exists(self.single_validation_video_save_dir):
            os.makedirs(self.single_validation_video_save_dir)
        _, _, self.fps = self.read_video_info(video_name=self.feature_filename)
        self.clf_name = os.path.basename(model_path).replace(".sav", "")
        self.video_path = self.find_video_of_file(self.video_dir, self.feature_filename)
        self.clf_data_save_path = os.path.join(
            self.clf_data_validation_dir, self.feature_filename + ".csv"
        )
        self.video_meta_data = get_video_meta_data(video_path=self.video_path)
        self.clf = read_df(file_path=model_path, file_type="pickle")
        self.in_df = read_df(feature_file_path, self.file_type)
        self.feature_file_path = feature_file_path
        self.temp_dir = os.path.join(self.single_validation_video_save_dir, "temp")
        self.video_save_path = os.path.join(
            self.single_validation_video_save_dir, self.feature_filename + ".mp4"
        )
        if not os.path.exists(self.temp_dir):
            os.makedirs(self.temp_dir)

    def __run_clf(self):
        self.prob_col_name = f"Probability_{self.clf_name}"
        self.in_df[self.prob_col_name] = self.clf_predict_proba(
            clf=self.clf,
            x_df=self.in_df,
            model_name=self.clf_name,
            data_path=self.feature_file_path,
        )
        self.in_df[self.clf_name] = np.where(
            self.in_df[self.prob_col_name] > self.discrimination_threshold, 1, 0
        )

    def __plug_bouts(self):
        self.data_df = plug_holes_shortest_bout(
            data_df=self.in_df,
            clf_name=self.clf_name,
            fps=self.fps,
            shortest_bout=self.shortest_bout,
        )

    def __save(self):
        write_df(
            df=self.data_df, file_type=self.file_type, save_path=self.clf_data_save_path
        )
        print(f"Predictions created for video {self.feature_filename}...")

    def __index_df_for_multiprocessing(self, data: list) -> list:
        for cnt, df in enumerate(data):
            df["group"] = cnt
        return data

    def __create_video(self):
        self.final_gantt_img = None
        self.bouts_df = None
        if self.create_gantt is not None:
            self.bouts_df = self.get_bouts_for_gantt(
                data_df=self.data_df, clf_name=self.clf_name, fps=self.fps
            )
            self.final_gantt_img = self.create_gantt_img(
                self.bouts_df,
                self.clf_name,
                len(self.data_df),
                self.fps,
                "Behavior gantt chart (entire session)",
            )
            self.final_gantt_img = self.resize_gantt(
                self.final_gantt_img, self.video_meta_data["height"]
            )

        if self.settings["styles"] is None:
            self.settings["styles"] = {}
            space_scaler, radius_scaler, resolution_scaler, font_scaler = (
                60,
                20,
                1500,
                1.5,
            )
            max_dim = max(self.video_meta_data["width"], self.video_meta_data["height"])
            self.settings["styles"]["circle size"] = int(
                radius_scaler / (resolution_scaler / max_dim)
            )
            self.settings["styles"]["font size"] = float(
                font_scaler / (resolution_scaler / max_dim)
            )
            self.settings["styles"]["space_scale"] = int(
                space_scaler / (resolution_scaler / max_dim)
            )

        data = np.array_split(self.in_df, self.cores)
        frm_per_core = data[0].shape[0]
        data = self.__index_df_for_multiprocessing(data=data)
        with multiprocessing.Pool(
            self.cores, maxtasksperchild=self.maxtasksperchild
        ) as pool:
            constants = functools.partial(
                self.validation_video_mp,
                bp_dict=self.animal_bp_dict,
                video_save_dir=self.temp_dir,
                settings=self.settings,
                video_meta_data=self.video_meta_data,
                video_path=self.video_path,
                gantt_setting=self.create_gantt,
                final_gantt=self.final_gantt_img,
                clf_data=self.data_df[self.clf_name].values,
                clrs=self.clr_lst,
                clf_name=self.clf_name,
                bouts_df=self.bouts_df,
            )
            try:
                for cnt, result in enumerate(
                    pool.imap(constants, data, chunksize=self.multiprocess_chunksize)
                ):
                    print(
                        "Image {}/{}, Video {}...".format(
                            str(int(frm_per_core * (result + 1))),
                            str(len(self.data_df)),
                            self.feature_filename,
                        )
                    )
                print(
                    "Joining {} multiprocessed video...".format(self.feature_filename)
                )
            except:
                pass
        concatenate_videos_in_folder(
            in_folder=self.temp_dir, save_path=self.video_save_path
        )
        self.timer.stop_timer()
        pool.terminate()
        pool.join()
        stdout_success(
            msg=f"Video {self.feature_filename} complete",
            elapsed_time=self.timer.elapsed_time_str,
        )

    def run(self):
        self.__run_clf()
        if self.shortest_bout > 1:
            self.__plug_bouts()
        else:
            self.data_df = self.in_df
        self.__save()
        self.__create_video()


# test = ValidateModelOneVideoMultiprocess(config_path=r'/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                              feature_file_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/features_extracted/Together_1.csv',
#                              model_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/models/generated_models/Attack.sav',
#                              discrimination_threshold=0.6,
#                              shortest_bout=50,
#                              cores=6,
#                              settings={'pose': True, 'animal_names': True, 'styles': None},
#                              create_gantt=None)
# test.run()



__author__ = "Simon Nilsson"

import functools
import multiprocessing
import os
import platform

import cv2
import numpy as np
import pandas as pd
from numba import jit, prange

import simba.mixins.plotting_mixin
from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.enums import Formats
from simba.utils.errors import NoSpecifiedOutputError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (concatenate_videos_in_folder, get_fn_ext,
                                    read_df, remove_a_folder)


def _heatmap_multiprocessor(
    data: np.array,
    video_setting: bool,
    frame_setting: bool,
    video_temp_dir: str,
    video_name: str,
    frame_dir: str,
    fps: int,
    style_attr: dict,
    max_scale: float,
    clf_name: str,
    aspect_ratio: float,
    size: tuple,
    make_clf_heatmap_plot: simba.mixins.plotting_mixin.PlottingMixin.make_clf_heatmap_plot,
):
    group = int(data[0][0][1])
    if video_setting:
        fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
        video_save_path = os.path.join(video_temp_dir, "{}.mp4".format(str(group)))
        video_writer = cv2.VideoWriter(video_save_path, fourcc, fps, size)

    for i in range(data.shape[0]):
        frame_id = int(data[i, 0, 0])
        frm_data = data[i, :, 2:]
        img = make_clf_heatmap_plot(
            frm_data=frm_data,
            max_scale=max_scale,
            palette=style_attr["palette"],
            aspect_ratio=aspect_ratio,
            shading=style_attr["shading"],
            clf_name=clf_name,
            img_size=size,
            final_img=False,
        )
        print(
            "Heatmap frame created: {}, Video: {}, Processing core: {}".format(
                str(frame_id + 1), video_name, str(group + 1)
            )
        )
        if video_setting:
            video_writer.write(img)

        if frame_setting:
            file_path = os.path.join(frame_dir, "{}.png".format(frame_id))
            cv2.imwrite(file_path, img)

    if video_setting:
        video_writer.release()

    return group


class HeatMapperClfMultiprocess(ConfigReader, PlottingMixin):
    """
    Create heatmaps representing the locations of the classified behavior.

    .. note::
       `GitHub visualizations tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/tutorial.md#step-11-visualizations>`__.

        .. image:: _static/img/heatmap.png
           :width: 300
           :align: center

    :param str config_path: path to SimBA project config file in Configparser format
    :param bool final_img_setting: If True, then create a single image representing the last frame of the input video
    :param bool video_setting: If True, then create a video of heatmaps.
    :param bool frame_setting: If True, then create individual heatmap frames.
    :param int bin_size: The rectangular size of each heatmap location in millimeters. For example, `50` will divide the video into
        5 centimeter rectangular spatial bins.
    :param str palette: Heatmap pallette. Eg. 'jet', 'magma', 'inferno','plasma', 'viridis', 'gnuplot2'
    :param str bodypart: The name of the body-part used to infer the location of the classified behavior
    :param str clf_name: The name of the classified behavior.
    :param int or 'auto' max_scale: The max value in the heatmap in seconds. E.g., with a value of `10`, if the classified behavior has occured
        >= 10 within a rectangular bins, it will be filled with the same color.
    :param int core_cnt: Number of cores to use.

    :examples:
    >>> heat_mapper_clf = HeatMapperClfMultiprocess(config_path='MyConfigPath', final_img_setting=False, video_setting=True, frame_setting=False, bin_size=50, palette='jet', bodypart='Nose_1', clf_name='Attack', max_scale=20)
    >>> heat_mapper_clf.create_heatmaps()
    """

    def __init__(
        self,
        config_path: str,
        final_img_setting: bool,
        video_setting: bool,
        frame_setting: bool,
        bodypart: str,
        clf_name: str,
        files_found: list,
        style_attr: dict,
        core_cnt: int,
    ):
        ConfigReader.__init__(self, config_path=config_path)
        PlottingMixin.__init__(self)

        if platform.system() == "Darwin":
            multiprocessing.set_start_method("spawn", force=True)
        if (not frame_setting) and (not video_setting) and (not final_img_setting):
            raise NoSpecifiedOutputError(
                msg="Please choose to select either heatmap videos, frames, and/or final image."
            )
        self.frame_setting, self.video_setting = frame_setting, video_setting
        self.final_img_setting, self.bp = final_img_setting, bodypart
        self.style_attr = style_attr
        self.bin_size, self.max_scale, self.palette, self.shading, self.core_cnt = (
            style_attr["bin_size"],
            style_attr["max_scale"],
            style_attr["palette"],
            style_attr["shading"],
            core_cnt,
        )
        self.clf_name, self.files_found = clf_name, files_found
        if not os.path.exists(self.heatmap_clf_location_dir):
            os.makedirs(self.heatmap_clf_location_dir)
        self.bp_lst = [self.bp + "_x", self.bp + "_y"]
        print("Processing {} video(s)...".format(str(len(self.files_found))))

    @staticmethod
    @jit(nopython=True)
    def __calculate_cum_array(clf_array: np.array, fps: int):
        cum_sum_arr = np.full(clf_array.shape, np.nan)
        for frm_idx in prange(clf_array.shape[0]):
            frame_cum_sum = np.full((clf_array.shape[1], clf_array.shape[2]), 0.0)
            sliced_arr = clf_array[0:frm_idx]
            for i in range(sliced_arr.shape[0]):
                for j in range(sliced_arr.shape[1]):
                    for k in range(sliced_arr.shape[2]):
                        frame_cum_sum[j][k] += sliced_arr[i][j][k]
            cum_sum_arr[frm_idx] = frame_cum_sum

        return cum_sum_arr / fps

    @staticmethod
    @jit(nopython=True)
    def __insert_group_idx_column(data: np.array, group: int, last_frm_idx: int):
        results = np.full((data.shape[0], data.shape[1], data.shape[2] + 2), np.nan)
        group_col = np.full((data.shape[1], 1), group)
        for frm_idx in prange(data.shape[0]):
            h_stack = np.hstack((group_col, data[frm_idx]))
            frm_col = np.full((h_stack.shape[0], 1), frm_idx + last_frm_idx)
            results[frm_idx] = np.hstack((frm_col, h_stack))

        return results

    def __calculate_bin_attr(
        self,
        data_df: pd.DataFrame,
        clf_name: str,
        bp_lst: list,
        px_per_mm: int,
        img_width: int,
        img_height: int,
        bin_size: int,
        fps: int,
    ):
        bin_size_px = int(float(px_per_mm) * float(bin_size))
        horizontal_bin_cnt = int(img_width / bin_size_px)
        vertical_bin_cnt = int(img_height / bin_size_px)
        aspect_ratio = round((vertical_bin_cnt / horizontal_bin_cnt), 3)

        clf_idx = (
            data_df[bp_lst][data_df[clf_name] == 1].reset_index().to_numpy().astype(int)
        )

        bin_dict = {}
        x_location, y_location = 0, 0
        for hbin in range(horizontal_bin_cnt):
            bin_dict[hbin] = {}
            for vbin in range(vertical_bin_cnt):
                bin_dict[hbin][vbin] = {
                    "top_left_x": x_location,
                    "top_left_y": y_location,
                    "bottom_right_x": x_location + bin_size_px,
                    "bottom_right_y": y_location + bin_size_px,
                }
                y_location += bin_size_px
            y_location = 0
            x_location += bin_size_px

        clf_array = np.zeros((len(data_df), vertical_bin_cnt, horizontal_bin_cnt))

        for clf_frame in clf_idx:
            for h_bin_name, v_dict in bin_dict.items():
                for v_bin_name, c in v_dict.items():
                    if (
                        clf_frame[1] <= c["bottom_right_x"]
                        and clf_frame[1] >= c["top_left_x"]
                    ):
                        if (
                            clf_frame[2] <= c["bottom_right_y"]
                            and clf_frame[2] >= c["top_left_y"]
                        ):
                            clf_array[int(clf_frame[0])][v_bin_name][h_bin_name] = 1

        clf_array = self.__calculate_cum_array(clf_array=clf_array, fps=fps)

        return clf_array, aspect_ratio

    def __calculate_max_scale(self, clf_array: np.array):
        return np.round(np.max(np.max(clf_array[-1], axis=0)), 3)

    def run(self):
        for file_cnt, file_path in enumerate(self.files_found):
            video_timer = SimbaTimer()
            video_timer.start_timer()
            _, self.video_name, _ = get_fn_ext(file_path)
            self.video_info, self.px_per_mm, self.fps = self.read_video_info(
                video_name=self.video_name
            )
            self.width, self.height = int(
                self.video_info["Resolution_width"].values[0]
            ), int(self.video_info["Resolution_height"].values[0])
            self.save_frame_folder_dir = os.path.join(
                self.heatmap_clf_location_dir, self.video_name + "_" + self.clf_name
            )
            self.video_folder = os.path.join(
                self.heatmap_clf_location_dir, self.video_name + "_" + self.clf_name
            )
            self.temp_folder = os.path.join(
                self.heatmap_clf_location_dir,
                self.video_name + "_" + self.clf_name,
                "temp",
            )
            if self.frame_setting:
                if os.path.exists(self.save_frame_folder_dir):
                    remove_a_folder(folder_dir=self.save_frame_folder_dir)
                if not os.path.exists(self.save_frame_folder_dir):
                    os.makedirs(self.save_frame_folder_dir)
            if self.video_setting:
                if os.path.exists(self.temp_folder):
                    remove_a_folder(folder_dir=self.temp_folder)
                    remove_a_folder(folder_dir=self.video_folder)
                os.makedirs(self.temp_folder)
                self.save_video_path = os.path.join(
                    self.heatmap_clf_location_dir,
                    "{}_{}.mp4".format(self.video_name, self.clf_name),
                )

            self.data_df = read_df(file_path=file_path, file_type=self.file_type)
            clf_array, aspect_ratio = self.__calculate_bin_attr(
                data_df=self.data_df,
                clf_name=self.clf_name,
                bp_lst=self.bp_lst,
                px_per_mm=self.px_per_mm,
                img_width=self.width,
                img_height=self.height,
                bin_size=self.bin_size,
                fps=self.fps,
            )

            if self.max_scale == "auto":
                self.max_scale = self.__calculate_max_scale(clf_array=clf_array)

            if self.final_img_setting:
                self.make_clf_heatmap_plot(
                    frm_data=clf_array[-1, :, :],
                    max_scale=self.max_scale,
                    palette=self.palette,
                    aspect_ratio=aspect_ratio,
                    file_name=os.path.join(
                        self.heatmap_clf_location_dir,
                        self.video_name + "_final_frm.png",
                    ),
                    shading=self.shading,
                    clf_name=self.clf_name,
                    img_size=(self.height, self.width),
                    final_img=True,
                )

            if self.video_setting or self.frame_setting:
                frame_arrays = np.array_split(clf_array, self.core_cnt)
                last_frm_idx = 0
                for frm_group in range(len(frame_arrays)):
                    split_arr = frame_arrays[frm_group]
                    frame_arrays[frm_group] = self.__insert_group_idx_column(
                        data=split_arr, group=frm_group, last_frm_idx=last_frm_idx
                    )
                    last_frm_idx = np.max(
                        frame_arrays[frm_group].reshape(
                            (frame_arrays[frm_group].shape[0], -1)
                        )
                    )
                frm_per_core = frame_arrays[0].shape[0]

                print(
                    "Creating heatmaps, multiprocessing (chunksize: {}, cores: {})...".format(
                        str(self.multiprocess_chunksize), str(self.core_cnt)
                    )
                )
                with multiprocessing.Pool(
                    self.core_cnt, maxtasksperchild=self.maxtasksperchild
                ) as pool:
                    constants = functools.partial(
                        _heatmap_multiprocessor,
                        video_setting=self.video_setting,
                        frame_setting=self.frame_setting,
                        style_attr=self.style_attr,
                        fps=self.fps,
                        video_temp_dir=self.temp_folder,
                        frame_dir=self.save_frame_folder_dir,
                        max_scale=self.max_scale,
                        aspect_ratio=aspect_ratio,
                        clf_name=self.clf_name,
                        size=(self.width, self.height),
                        video_name=self.video_name,
                        make_clf_heatmap_plot=self.make_clf_heatmap_plot,
                    )

                    for cnt, result in enumerate(
                        pool.imap(
                            constants,
                            frame_arrays,
                            chunksize=self.multiprocess_chunksize,
                        )
                    ):
                        print(
                            "Image {}/{}, Video {}/{}...".format(
                                str(int(frm_per_core * (result + 1))),
                                str(len(self.data_df)),
                                str(file_cnt + 1),
                                str(len(self.files_found)),
                            )
                        )
                    pool.terminate()
                    pool.join()

                if self.video_setting:
                    print("Joining {} multiprocessed video...".format(self.video_name))
                    concatenate_videos_in_folder(
                        in_folder=self.temp_folder, save_path=self.save_video_path
                    )

                video_timer.stop_timer()
                print(
                    "Heatmap video {} complete (elapsed time: {}s) ...".format(
                        self.video_name, video_timer.elapsed_time_str
                    )
                )

        self.timer.stop_timer()
        stdout_success(
            msg="heatmap visualizations for {} videos created in project_folder/frames/output/heatmap_classifier locations directory",
            elapsed_time=self.timer.elapsed_time_str,
        )


# test = HeatMapperClfMultiprocess(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                      style_attr = {'palette': 'jet', 'shading': 'gouraud', 'bin_size': 100, 'max_scale': 'auto'},
#                      final_img_setting=False,
#                      video_setting=True,
#                      frame_setting=False,
#                      bodypart='Nose_1',
#                      clf_name='Attack',
#                                core_cnt=5,
#                      files_found=['/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/machine_results/Together_1.csv'])
# test.create_heatmaps()



__author__ = "Simon Nilsson"

from tkinter import *

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import Entry_Box
from simba.utils.checks import check_str
from simba.utils.read_write import archive_processed_files


class ArchiveProcessedFilesPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str):
        PopUpMixin.__init__(self, title="ADD CLASSIFIER")
        ConfigReader.__init__(self, config_path=config_path)
        self.archive_eb = Entry_Box(self.main_frm, "ARCHIVE DIRECTORY NAME", "25")
        archive_btn = Button(
            self.main_frm, text="RUN ARCHIVE", fg="blue", command=lambda: self.run()
        )
        self.archive_eb.grid(row=0, column=0, sticky=NW)
        archive_btn.grid(row=1, column=0, sticky=NW)

    def run(self):
        archive_name = self.archive_eb.entry_get.strip()
        check_str(name="CLASSIFIER NAME", value=archive_name)
        archive_processed_files(config_path=self.config_path, archive_name=archive_name)



__author__ = "Simon Nilsson"

import os

from simba.mixins.config_reader import ConfigReader
from simba.roi_tools.ROI_analyzer import ROIAnalyzer
from simba.utils.printing import stdout_success


class ROIMovementAnalyzer(ConfigReader):
    """

    Compute movements of individual animals within user-defined ROIs.

    :param str config_path: Path to SimBA project config file in Configparser format

    .. note::
       `ROI tutorials <https://github.com/sgoldenlab/simba/blob/master/docs/ROI_tutorial_new.md>`__.

    Examples
    ----------
    >>> _ = ROIMovementAnalyzer(config_path='MyProjectConfig')
    """

    def __init__(self, config_path: str):
        ConfigReader.__init__(self, config_path=config_path)
        self.roi_analyzer = ROIAnalyzer(
            ini_path=config_path,
            data_path="outlier_corrected_movement_location",
            calculate_distances=True,
        )
        self.roi_analyzer.run()
        self.save_data()

    def save_data(self):
        """
        Save ROI movement analysis. Results are stored in the ``project_folder/logs`` directory
        of the SimBA project.

        Returns
        ----------
        None
        """
        save_path = os.path.join(
            self.roi_analyzer.logs_path,
            "ROI_movement_data_" + self.roi_analyzer.datetime + ".csv",
        )
        self.roi_analyzer.movements_df.to_csv(save_path)
        self.timer.stop_timer()
        stdout_success(
            'ROI movement data saved in the "project_folder/logs/" directory',
            elapsed_time=self.timer.elapsed_time_str,
        )


# test = ROIMovementAnalyzer(config_path='/Users/simon/Desktop/envs/troubleshooting/two_animals_16bp_032023/project_folder/project_config.ini')



import os
from tkinter import *

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.plotting.ez_lineplot import draw_line_plot
from simba.utils.errors import NoFilesFoundError
from simba.utils.read_write import find_all_videos_in_directory


class QuickLineplotPopup(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str):
        PopUpMixin.__init__(self, title="SIMPLE LINE PLOT")
        ConfigReader.__init__(self, config_path=config_path)
        video_filepaths = find_all_videos_in_directory(
            directory=os.path.join(self.project_path, "videos")
        )
        self.video_files = [os.path.basename(x) for x in video_filepaths]
        if len(self.video_files) == 0:
            raise NoFilesFoundError(
                msg="SIMBA ERROR: No files detected in the project_folder/videos directory.",
                source=self.__class__.__name__,
            )

        self.all_body_parts = []
        for animal, bp_cords in self.animal_bp_dict.items():
            for bp_dim, bp_data in bp_cords.items():
                self.all_body_parts.extend(([x[:-2] for x in bp_data]))

        self.settings_frm = LabelFrame(self.main_frm, text="Settings")
        self.settings_frm.grid(row=0, column=0, sticky=NW)

        self.video_lbl = Label(self.settings_frm, text="Video: ")
        self.chosen_video_val = StringVar(value=self.video_files[0])
        self.chosen_video_dropdown = OptionMenu(
            self.settings_frm, self.chosen_video_val, *self.video_files
        )
        self.video_lbl.grid(row=0, column=0, sticky=NW)
        self.chosen_video_dropdown.grid(row=0, column=1, sticky=NW)

        self.body_part_lbl = Label(self.settings_frm, text="Body-part: ")
        self.chosen_bp_val = StringVar(value=self.all_body_parts[0])
        self.choose_bp_dropdown = OptionMenu(
            self.settings_frm, self.chosen_bp_val, *self.all_body_parts
        )
        self.body_part_lbl.grid(row=1, column=0, sticky=NW)
        self.choose_bp_dropdown.grid(row=1, column=1, sticky=NW)

        run_btn = Button(
            self.settings_frm,
            text="Create path plot",
            command=lambda: draw_line_plot(
                self.config_path, self.chosen_video_val.get(), self.chosen_bp_val.get()
            ),
        )
        run_btn.grid(row=2, column=1, pady=10)
        self.main_frm.mainloop()


# _ = QuickLineplotPopup(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')



__author__ = "Simon Nilsson"

import os
import shutil
from tkinter import *
from typing import Optional

import pandas as pd

import simba
from simba.utils.checks import check_file_exist_and_readable
from simba.utils.lookups import get_bp_config_codes
from simba.utils.printing import stdout_trash


class PoseResetter(object):
    """
    Launch GUI for deleting all **user-defined** pose-estimation schematics, diagrams and other settings from the
    SimBA installation.

    :param Optional[TopLevel] master: Tkinter TopLevel window. Default: None.

    Examples
    ----------
    >>> _ = PoseResetter(master=None)
    """

    def __init__(self, master: Optional[Toplevel] = None):
        self.default_pose_configs_cnt = len(get_bp_config_codes().keys())
        self.simba_dir = os.path.dirname(simba.__file__)
        self.pose_configs_dir = os.path.join(self.simba_dir, "pose_configurations")
        self.archive_dir = os.path.join(self.simba_dir, "pose_configurations_archive")
        if not os.path.exists(self.archive_dir):
            os.makedirs(self.archive_dir)
        self.archive_folders_cnt = len(os.listdir(self.archive_dir))
        self.archive_subdir = os.path.join(
            self.archive_dir,
            "pose_configurations_archive_{}".format(str(self.archive_folders_cnt + 1)),
        )
        shutil.copytree(self.pose_configs_dir, self.archive_subdir)

        self.bp_names_csv_path = os.path.join(
            self.pose_configs_dir, "bp_names", "bp_names.csv"
        )
        self.pose_config_names_csv_path = os.path.join(
            self.pose_configs_dir, "configuration_names", "pose_config_names.csv"
        )
        self.no_animals_csv_path = os.path.join(
            self.pose_configs_dir, "no_animals", "no_animals.csv"
        )
        self.schematics_path = os.path.join(self.pose_configs_dir, "schematics")

        for file_path in [
            self.bp_names_csv_path,
            self.pose_config_names_csv_path,
            self.no_animals_csv_path,
            self.no_animals_csv_path,
        ]:
            check_file_exist_and_readable(file_path=file_path)
            df = pd.read_csv(file_path, header=None, error_bad_lines=False)
            df = df.iloc[0 : self.default_pose_configs_cnt]
            df.to_csv(file_path, index=False, header=False)

        default_pic_list = []
        user_pic_lst = os.listdir(self.schematics_path)
        for idx in range(self.default_pose_configs_cnt):
            default_pic_list.append("{}.png".format(str(idx + 1)))
        for i in list(set(user_pic_lst) - set(default_pic_list)):
            os.remove(os.path.join(self.schematics_path, i))

        stdout_trash(
            msg="User-defined pose-estimation configuration reset. User-defined poses removed."
        )
        master.destroy()


# test = PoseResetter()



from pygifsicle import optimize

pt = optimize("/Users/simon/Downloads/Studio_Project (3).gif")


# # import cv2
# # import numpy as np
# # import os
# #
# # IMG_PATH = '/Users/simon/Desktop/splash.png'
# # OUT_PATH = '/Users/simon/Desktop/envs/simba_dev/splash'
# # alpha_lst = np.linspace(0, 1, 25).tolist()
# # for i in range(25):
# #     alpha_lst.append(1)
# #
# # #alpha_lst = alpha_lst[0:1]
# #
# # for cnt, i in enumerate(alpha_lst):
# #     img = cv2.imread(IMG_PATH)
# #     bgra = cv2.cvtColor(img, cv2.COLOR_BGR2BGRA)
# #     bgra[..., 3] = int(254 * i)
# #     cv2.imwrite(os.path.join(OUT_PATH, 'image_' + str(cnt) + '.png'), bgra)
# #
#
#
#
# import glob
# import contextlib
# from PIL import Image
#
# # filepaths
# fp_in = "/Users/simon/Desktop/envs/simba_dev/splash/image_*.png"
# fp_out = "/Users/simon/Desktop/envs/simba_dev/splash_gif.gif"
#
# # use exit stack to automatically close opened images
# with contextlib.ExitStack() as stack:
#
#     # lazily load images
#     imgs = (stack.enter_context(Image.open(f))
#             for f in sorted(glob.glob(fp_in)))
#     print(imgs)
#
#     # extract  first image from iterator
#     img = next(imgs)
#
#     # https://pillow.readthedocs.io/en/stable/handbook/image-file-formats.html#gif
#     img.save(fp=fp_out, format='GIF', append_images=imgs,
#              save_all=True, duration=2500, loop=1)
#
#
#
#
# frames = [Image.open(image) for image in glob.glob(f"{'/Users/simon/Desktop/envs/simba_dev/splash/'}/*")]
# frame_one = frames[0]
# frame_one.save("download.gif", format="GIF", append_images=frames, save_all=True, duration=1000, loop=1)



import os
from configparser import ConfigParser
from tkinter import *

import pandas as pd

from simba.utils.enums import ConfigKey, Formats, Keys, Paths
from simba.utils.errors import NoROIDataError
from simba.utils.printing import stdout_trash
from simba.utils.read_write import get_fn_ext, read_config_file


def reset_video_ROIs(config_path, filename):
    _, file_name_wo_ext, VideoExtension = get_fn_ext(filename)
    config = ConfigParser()
    configFile = str(config_path)
    config.read(configFile)
    vidInfPath = config.get(
        ConfigKey.GENERAL_SETTINGS.value, ConfigKey.PROJECT_PATH.value
    )
    logFolderPath = os.path.join(vidInfPath, "logs")
    ROIcoordinatesPath = os.path.join(logFolderPath, Paths.ROI_DEFINITIONS.value)
    if not os.path.isfile(ROIcoordinatesPath):
        raise NoROIDataError(
            msg="Cannot delete ROI definitions: no definitions exist to delete"
        )

    else:
        rectanglesInfo = pd.read_hdf(ROIcoordinatesPath, key=Keys.ROI_RECTANGLES.value)
        circleInfo = pd.read_hdf(ROIcoordinatesPath, key=Keys.ROI_CIRCLES.value)
        polygonInfo = pd.read_hdf(ROIcoordinatesPath, key=Keys.ROI_POLYGONS.value)
        store = pd.HDFStore(ROIcoordinatesPath, mode="w")

    try:
        rectanglesInfo = rectanglesInfo[rectanglesInfo["Video"] != file_name_wo_ext]
    except KeyError:
        pass
    store["rectangles"] = rectanglesInfo

    try:
        circleInfo = circleInfo[circleInfo["Video"] != file_name_wo_ext]
    except KeyError:
        pass
    store["circleDf"] = circleInfo

    try:
        polygonInfo = polygonInfo[polygonInfo["Video"] != file_name_wo_ext]
    except KeyError:
        pass
    store["polygons"] = polygonInfo

    print("Deleted ROI record: " + str(file_name_wo_ext))
    store.close()


def delete_all_ROIs(config_path: str):
    def delete_file(config_path):
        config = read_config_file(config_path=config_path)
        project_path = config.get(
            ConfigKey.GENERAL_SETTINGS.value, ConfigKey.PROJECT_PATH.value
        )
        roi_data_path = os.path.join(project_path, "logs", Paths.ROI_DEFINITIONS.value)

        if not os.path.isfile(roi_data_path):
            raise NoROIDataError(
                msg=f"No ROI definitions exist in this SimBA project. Expected file at path {roi_data_path}"
            )
        else:
            os.remove(roi_data_path)
            close_window()
            stdout_trash(
                msg=f"SIMBA COMPLETE: All ROI definitions deleted in this SimBA project ({roi_data_path})"
            )

    def close_window():
        delete_confirm_win.destroy()
        delete_confirm_win.update()

    delete_confirm_win = Toplevel()
    delete_confirm_win.minsize(200, 200)

    question_frame = LabelFrame(
        delete_confirm_win, text="Confirm", font=("Arial", 16, "bold"), padx=5, pady=5
    )
    question_lbl = Label(
        question_frame,
        text="Do you want to delete all defined ROIs in the project?",
        font=Formats.LABELFRAME_HEADER_FORMAT.value,
    )

    yes_button = Button(
        question_frame, text="YES", fg="black", command=lambda: delete_file(config_path)
    )
    no_button = Button(
        question_frame, text="NO", fg="black", command=lambda: close_window()
    )

    question_frame.grid(row=0, sticky=W)
    question_lbl.grid(row=1, column=0, sticky=W, pady=10, padx=10)
    yes_button.grid(row=2, column=1, sticky=W, pady=10, padx=10)
    no_button.grid(row=2, column=2, sticky=W, pady=10, padx=10)



__author__ = "Simon Nilsson"

import os
from typing import Dict, List

import cv2
import numpy as np
import pandas as pd
from joblib import Parallel, delayed

from simba.data_processors.movement_calculator import MovementCalculator
from simba.mixins.config_reader import ConfigReader
from simba.utils.enums import Formats
from simba.utils.errors import NoSpecifiedOutputError
from simba.utils.lookups import get_color_dict
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import get_fn_ext


class DataPlotter(ConfigReader):
    """
    Tabular data visualization of animal movement and distances in the current frame and their aggregate
    statistics.

    :parameter str config_path: path to SimBA project config file in Configparser format

    .. note::
        `Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/Scenario2.md#visualizing-data-tables`>_.

    .. image:: _static/img/data_plot.png
       :width: 300
       :align: center

    :examples:
    >>> _ = DataPlotter(config_path='MyConfigPath').run()
    """

    def __init__(
        self,
        config_path: str,
        style_attr: Dict,
        body_part_attr: List[List[str]],
        data_paths: List[str],
        video_setting: bool,
        frame_setting: bool,
    ):
        super().__init__(config_path=config_path)
        if (not video_setting) and (not frame_setting):
            raise NoSpecifiedOutputError(
                msg="SIMBA ERROR: Please choose to create video and/or frames data plots. SimBA found that you ticked neither video and/or frames"
            )
        self.video_setting, self.frame_setting = video_setting, frame_setting
        self.files_found, self.style_attr, self.body_part_attr = (
            data_paths,
            style_attr,
            body_part_attr,
        )
        if not os.path.exists(self.data_table_path):
            os.makedirs(self.data_table_path)
        self.process_movement()
        print(f"Processing {len(self.files_found)} video(s)...")

    def __compute_spacings(self):
        """
        Private helper to compute appropriate spacing between printed text.
        """
        self.loc_dict = {}
        self.loc_dict["Animal"] = (50, 20)
        self.loc_dict["total_movement_header"] = (250, 20)
        self.loc_dict["current_velocity_header"] = (475, 20)
        self.loc_dict["animals"] = {}
        y_cord, x_cord = 75, 15
        for animal_cnt, animal_name in enumerate(self.video_data.columns):
            self.loc_dict["animals"][animal_name] = {}
            self.loc_dict["animals"][animal_name]["index_loc"] = (50, y_cord)
            self.loc_dict["animals"][animal_name]["total_movement_loc"] = (250, y_cord)
            self.loc_dict["animals"][animal_name]["current_velocity_loc"] = (
                475,
                y_cord,
            )
            y_cord += 50

    def process_movement(self):
        movement_processor = MovementCalculator(
            config_path=self.config_path,
            file_paths=self.files_found,
            threshold=0.00,
            body_parts=[x[0] for x in self.body_part_attr],
        )
        movement_processor.run()
        self.movement = movement_processor.movement_dfs

    def run(self):
        def multiprocess_img_creation(
            video_data_slice: list,
            location_dict: dict,
            animal_ids: list,
            video_data: pd.DataFrame,
            style_attr: dict,
            body_part_attr: dict,
        ):
            color_dict = get_color_dict()
            img = np.zeros((style_attr["size"][1], style_attr["size"][0], 3))
            img[:] = color_dict[style_attr["bg_color"]]
            cv2.putText(
                img,
                "Animal",
                location_dict["Animal"],
                cv2.FONT_HERSHEY_TRIPLEX,
                0.5,
                color_dict[style_attr["header_color"]],
                style_attr["font_thickness"],
            )
            cv2.putText(
                img,
                "Total movement (cm)",
                location_dict["total_movement_header"],
                cv2.FONT_HERSHEY_TRIPLEX,
                0.5,
                color_dict[style_attr["header_color"]],
                style_attr["font_thickness"],
            )
            cv2.putText(
                img,
                "Velocity (cm/s)",
                location_dict["current_velocity_header"],
                cv2.FONT_HERSHEY_TRIPLEX,
                0.5,
                color_dict[style_attr["header_color"]],
                style_attr["font_thickness"],
            )
            for animal_cnt, animal_name in enumerate(animal_ids):
                clr = color_dict[body_part_attr[animal_cnt][1]]
                total_movement = str(
                    round(
                        video_data[animal_name]
                        .iloc[0 : video_data_slice.index.max()]
                        .sum()
                        / 10,
                        style_attr["data_accuracy"],
                    )
                )
                current_velocity = str(
                    round(
                        video_data_slice[animal_name].sum() / 10,
                        style_attr["data_accuracy"],
                    )
                )
                cv2.putText(
                    img,
                    animal_name,
                    location_dict["animals"][animal_name]["index_loc"],
                    cv2.FONT_HERSHEY_TRIPLEX,
                    0.5,
                    clr,
                    1,
                )
                cv2.putText(
                    img,
                    total_movement,
                    location_dict["animals"][animal_name]["total_movement_loc"],
                    cv2.FONT_HERSHEY_TRIPLEX,
                    0.5,
                    clr,
                    1,
                )
                cv2.putText(
                    img,
                    current_velocity,
                    location_dict["animals"][animal_name]["current_velocity_loc"],
                    cv2.FONT_HERSHEY_TRIPLEX,
                    0.5,
                    clr,
                    1,
                )
            return img

        for file_cnt, file_path in enumerate(self.files_found):
            video_timer = SimbaTimer(start=True)
            _, video_name, _ = get_fn_ext(file_path)
            self.video_data = pd.DataFrame(self.movement[video_name])
            self.__compute_spacings()
            _, _, self.fps = self.read_video_info(video_name=video_name)
            if self.video_setting:
                self.fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
                self.video_save_path = os.path.join(
                    self.data_table_path, video_name + ".mp4"
                )
                self.writer = cv2.VideoWriter(
                    self.video_save_path, self.fourcc, self.fps, self.style_attr["size"]
                )
            if self.frame_setting:
                self.frame_save_path = os.path.join(self.data_table_path, video_name)
                if not os.path.exists(self.frame_save_path):
                    os.makedirs(self.frame_save_path)
            video_data_lst = np.array_split(
                pd.DataFrame(self.video_data), int(len(self.video_data) / self.fps)
            )
            self.imgs = Parallel(
                n_jobs=self.cpu_to_use, verbose=1, backend="threading"
            )(
                delayed(multiprocess_img_creation)(
                    x,
                    self.loc_dict,
                    self.video_data.columns,
                    self.video_data,
                    self.style_attr,
                    self.body_part_attr,
                )
                for x in video_data_lst
            )
            frm_cnt = 0
            for img_cnt, img in enumerate(self.imgs):
                for frame_cnt in range(int(self.fps)):
                    if self.video_setting:
                        self.writer.write(np.uint8(img))
                    if self.frame_setting:
                        frm_save_name = os.path.join(
                            self.frame_save_path, "{}.png".format(str(frm_cnt))
                        )
                        cv2.imwrite(frm_save_name, np.uint8(img))
                    frm_cnt += 1
                    print(
                        "Frame: {} / {}. Video: {} ({}/{})".format(
                            str(frm_cnt),
                            str(len(self.video_data)),
                            video_name,
                            str(file_cnt + 1),
                            len(self.files_found),
                        )
                    )

            print("Data tables created for video {}...".format(video_name))
            if self.video_setting:
                self.writer.release()
                video_timer.stop_timer()
                print(
                    "Video {} complete (elapsed time {}s)...".format(
                        video_name, video_timer.elapsed_time_str
                    )
                )

        self.timer.stop_timer()
        stdout_success(
            msg=f"All data table videos created inside {self.data_table_path}",
            elapsed_time=self.timer.elapsed_time_str,
        )


# style_attr = {'bg_color': 'White', 'header_color': 'Black', 'font_thickness': 1, 'size': (640, 480), 'data_accuracy': 2}
# body_part_attr = [['Ear_left_1', 'Grey'], ['Ear_right_2', 'Red']]
# data_paths = ['/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/machine_results/Together_1.csv']
#
#
# test = DataPlotter(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                    style_attr=style_attr,
#                    body_part_attr=body_part_attr,
#                    data_paths=data_paths,
#                    video_setting=True,
#                    frame_setting=False)
# test.run()

# style_attr = {'bg_color': 'White', 'header_color': 'Black', 'font_thickness': 1, 'size': (640, 480), 'data_accuracy': 2}
# body_part_attr = [['Ear_left_1', 'Grey'], ['Ear_right_2', 'Red']]
# data_paths = ['/Users/simon/Desktop/envs/simba_dev/tests/test_data/two_C57_madlc/project_folder/csv/outlier_corrected_movement_location/Together_1.csv']
#
#
# test = DataPlotter(config_path='/Users/simon/Desktop/envs/simba_dev/tests/test_data/two_C57_madlc/project_folder/project_config.ini',
#                    style_attr=style_attr,
#                    body_part_attr=body_part_attr,
#                    data_paths=data_paths,
#                    video_setting=True,
#                    frame_setting=False)
# test.create_data_plots()
